"""
monobit.plumbing.help - contextual help

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from ..storage.base import (
    loaders, savers,
    container_loaders, container_savers,
    encoders, decoders, containers,
)
from .args import GLOBAL_ARG_PREFIX, ARG_PREFIX, FALSE_PREFIX


# doc string alignment in usage text
HELP_TAB = 25


def get_argdoc(func, for_arg):
    """Get documentation for function argument."""
    if not func.__doc__:
        return ''
    for line in func.__doc__.splitlines():
        line = line.strip()
        if not line or ':' not in line:
            continue
        arg, doc = line.split(':', 1)
        if arg.strip() == for_arg:
            return doc.strip()
    return ''


def get_funcdoc(func):
    """Get documentation for function."""
    if not func.__doc__:
        return ''
    for line in func.__doc__.splitlines():
        line = line.strip()
        if line:
            return line
    return ''


def _print_option_help(name, vartype, doc, tab, prefix, *, add_unsetter=True):
    name = name.replace('_', '-')
    if vartype == bool:
        print(f'{prefix}{name}\t{doc}'.expandtabs(tab))
        if add_unsetter:
            print(f'{prefix}{FALSE_PREFIX}{name}\tunset {prefix}{name}'.expandtabs(tab))
    else:
        print(f'{prefix}{name}=...\t{doc}'.expandtabs(tab))


def _print_with_bar(name, doc):
    print(f'{name} '.ljust(HELP_TAB-1, '-') + f' {doc}')


def _print_section(section_name, func):
    _print_with_bar(section_name, get_funcdoc(func))
    for name, vartype in func.__annotations__.items():
        doc = get_argdoc(func, name)
        _print_option_help(name, vartype, doc, HELP_TAB, ARG_PREFIX)
    print()



def print_help(command_args, usage, operations, global_options):
    print(usage)
    print()
    print('Options')
    print('=======')
    print()
    for name, (vartype, doc) in global_options.items():
        _print_option_help(name, vartype, doc, HELP_TAB, GLOBAL_ARG_PREFIX, add_unsetter=False)

    if not command_args or len(command_args) == 1 and not command_args[0].command:
        print()
        print('Commands')
        print('========')
        print()
        for op, func in operations.items():
            _print_with_bar(op, get_funcdoc(func))
    else:
        print()
        print('Commands and their options')
        print('==========================')
        print()
        for ns in command_args:
            op = ns.command
            if not op:
                continue
            func = ns.func
            _print_section(op, func)
            if op in ('load', 'save', 'to'):
                _print_context_help(**vars(ns))


def _print_context_help(command, args, kwargs, **ignore):
    format = kwargs.get('format', '')
    if command == 'load':
        funcs = container_loaders.get_for(format=format)
        if not funcs:
            funcs = loaders.get_for(format=format)
    else:
        funcs = container_savers.get_for(format=format)
        if not funcs:
            funcs = savers.get_for(format=format)
    if funcs:
        func = funcs[0]
        _print_section(f'{command} -format={func.format}', func)
    container_format = kwargs.get('container_format', '')
    if command == 'load':
        funcs = decoders.get_for(format=container_format)
    else:
        funcs = encoders.get_for(format=container_format)
    if funcs:
        func = funcs[0]
        _print_section(f'-container-format={func.format}', func)
    container_classes = containers.get_for(format=container_format)
    if containers:
        if command == 'load':
            func = container_classes[0].decode
        else:
            func = container_classes[0].encode
        _print_section(f'-container-format={container_format}', func)



"""
monobit.storage.fontformats.vtfont - FreeBSD vt consolve font file

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from collections import deque

from monobit.storage import loaders, savers, Magic
from monobit.core import Glyph, Font, Char, Raster
from monobit.base.struct import big_endian as be
from monobit.base.binary import ceildiv

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single, make_contiguous, ensure_charcell


# https://codeberg.org/FreeBSD/freebsd-src/src/branch/main/sys/sys/font.h

# see also John Zaitseff's freebsdvfntio
# https://www.zap.org.au/projects/console-fonts-utils/src/freebsdvfntio.py


_VFNT_MAGIC = b'VFNT0002'

_VFNT_HEADER = be.Struct(
    fh_magic='8s',
    fh_width='uint8',
    fh_height='uint8',
    fh_pad='uint16',
    fh_glyph_count='uint32',
    fh_map_count=be.uint32 * 4,
)

# > Fonts support normal and bold weights, and single and double width glyphs.
# > Mapping tables are used to map Unicode points to glyphs.  They are sorted by
# > code point, and vtfont_lookup() uses this to perform a binary search.  Each
# > font has four mapping tables: two weights times two halves (left/single,
# > right).  When a character is not present in a bold map the glyph from the
# > normal map is used.  When no glyph is available, it uses glyph 0, which is
# > normally equal to U+FFFD.

_VFNT_MAP = be.Struct(
    # starting unicode code point
    vfm_src='uint32',
    # starting glyph index
    vfm_dst='uint16',
    # number of glyphs to apply to, less one
    vfm_len='uint16',
)


@loaders.register(
    name='vtfont',
    magic=(_VFNT_MAGIC,),
    patterns=('*.fnt',),
)
def load_vfnt(instream):
    """Load a vtfont font."""
    header = _VFNT_HEADER.read_from(instream)
    glyphs = load_bitmap(
        instream,
        width=header.fh_width, height=header.fh_height,
        count=header.fh_glyph_count,
    ).glyphs
    # generated codepoint is not meaningful
    raw_glyphs = list(_g.modify(codepoint=None) for _g in glyphs)
    # > VFNT_MAP_NORMAL = 0,    /* Normal font. */
    # > VFNT_MAP_NORMAL_RIGHT,  /* Normal font right hand. */
    # > VFNT_MAP_BOLD,          /* Bold font. */
    # > VFNT_MAP_BOLD_RIGHT,    /* Bold font right hand. */
    rle_maps = (
        (_VFNT_MAP * header.fh_map_count[_i]).read_from(instream)
        for _i in range(4)
    )
    # label glyph halves
    expanded_maps = {}, {}, {}, {}
    for map, exp_map in zip(rle_maps, expanded_maps):
        for entry in map:
            for offset in range(entry.vfm_len+1):
                char = Char(chr(entry.vfm_src + offset))
                index = entry.vfm_dst + offset
                exp_map[char] = index
    # construct dict to hold original and merged rasters
    rasters = {
        _index: _glyph.pixels
        for _index, _glyph in enumerate(raw_glyphs)
    }
    for map_index in (0, 2):
        # merge left- and right-hand halves
        for char, rh_index in expanded_maps[map_index+1].items():
            # use index 0 as default (per comments in the source)
            lh_index = expanded_maps[map_index].get(char, 0)
            expanded_maps[map_index][char] = (lh_index, rh_index)
            if (lh_index, rh_index) not in rasters:
                rasters[(lh_index, rh_index)] = Raster.concatenate(
                    rasters[lh_index], rasters[rh_index]
                )
    # create index-to-label maps
    index_to_labels = {0: {}, 2: {}}
    for map_index in (0, 2):
        for char, indices in expanded_maps[map_index].items():
            if indices in index_to_labels[map_index]:
                index_to_labels[map_index][indices].append(char)
            else:
                index_to_labels[map_index][indices] = [char]
    # create labelled glyphs
    normal_glyphs = tuple(
        Glyph(rasters[_index], labels=_labels)
        for _index, _labels in index_to_labels[0].items()
    )
    bold_glyphs = tuple(
        Glyph(rasters[_index], labels=_labels)
        for _index, _labels in index_to_labels[2].items()
    )
    # create fonts for normal and bold, if present
    fonts = tuple(
        Font(_g, weight=_w)
        for _w, _g in {'regular': normal_glyphs, 'bold': bold_glyphs}.items()
        if _g
    )
    return fonts


@savers.register(linked=load_vfnt)
def save_vfnt(fonts, outstream):
    if len(fonts) > 2 or len(fonts) == 2 and (
            fonts[0].cell_size != fonts[1].cell_size
        ):
        raise ValueError(
            'This format can store only two fonts (regular and bold weight) '
            'of matching cell size.'
        )
    # don't enforce different weights, but if they do have them, put in order
    if fonts[0].weight == 'bold' and fonts[1].weight == 'regular':
        fonts = reversed(fonts)
    # select glyphs with char labels only
    fonts = tuple(_f.label().subset(chars=_f.get_chars()) for _f in fonts)
    # split double-width glyphs
    split_fonts = []
    for font in fonts:
        font = font.equalise_horizontal()
        if font.spacing == 'multi-cell':
            left = Font(
                _g if _g.width == font.cell_size.x
                else _g.crop(right=font.cell_size.x, adjust_metrics=False)
                for _g in font.glyphs
                if _g.width <= font.cell_size.x * 2
            )
            right = Font(
                _g.crop(left=font.cell_size.x, adjust_metrics=False)
                for _g in font.glyphs
                if _g.width == font.cell_size.x * 2
            )
            split_fonts.extend((left, right))
        elif font.spacing == 'character-cell':
            split_fonts.extend((font, Font()))
        else:
            raise ValueError(
                'This format only supports character-cell and multi-cell fonts.'
            )
    fonts = split_fonts
    maps = []
    base_index = 0
    for font in fonts:
        # generate char ranges
        # use first char label initially, assumption is that these are likely contiguous
        ucps = [
            # skip unicode sequences
            (_i, deque(ord(_c) for _c in _g.chars if len(_c) == 1))
            for _i, _g in enumerate(font.glyphs, base_index)
        ]
        map = []
        while any(_deq for _i, _deq in ucps):
            last_char = -2
            current_entry = None
            for index, chardeq in ucps:
                if not chardeq:
                    last_char = -2
                    continue
                char = chardeq.popleft()
                if char != last_char + 1 or current_entry is None:
                    if current_entry is not None:
                        map.append(current_entry)
                    current_entry = _VFNT_MAP(vfm_src=char, vfm_dst=index, vfm_len=0)
                else:
                    current_entry.vfm_len += 1
                last_char = char
            map.append(current_entry)
        maps.append(map)
        base_index += len(font.glyphs)
    # generate header
    header = _VFNT_HEADER(
        fh_magic=_VFNT_MAGIC,
        fh_width=fonts[0].cell_size.x,
        fh_height=fonts[0].cell_size.y,
        fh_pad=0,
        fh_glyph_count=sum(len(_f.glyphs) for _f in fonts),
        fh_map_count=tuple(len(_m) for _m in maps),
    )
    outstream.write(bytes(header))
    for font in fonts:
        save_bitmap(outstream, font)
    for map in maps:
        map = (_VFNT_MAP * len(map))(*map)
        outstream.write(bytes(map))



"""
monobit.base - supporting classes

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .basetypes import *
from .properties import reverse_dict, extend_string, Props
from .cachedprops import HasProps, checked_property, writable_property
from . import struct
from . import binary
from . import blocks
from .imports import import_all, safe_import



"""
monobit.storage - recognise files, traverse filesystems, load and save fonts

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .base import loaders, savers
from .fontfiles import load, save
from .magic import Regex, Glob, Magic
from .streams import Stream, KeepOpen, get_stringio, get_bytesio
from . import streams


# ensure plugins get loaded
from . import containerformats as _containerformats
from . import fontformats as _fontformats
from . import wrapperformats as _wrapperformats



"""
monobit.storage.fontformats.raw.drhalo - Dr. Halo / Dr. Genius font files

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, Magic
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap


###############################################################################
# Dr. Halo / Dr. Genius F*X*.FON
# this covers the bitmap fonts only
# most fonts included with Dr Halo are vector fonts
# for which I have not yet deciphered the encoding


_DRHALO_SIG = b'AH'

@loaders.register(
    name='drhalo',
    magic=(_DRHALO_SIG,),
    patterns=('*.fon',),
)
def load_drhalo(instream, first_codepoint:int=0):
    """Load a Dr Halo / Dr Genius .FON bitmap font."""
    start = instream.read(16)
    if not start.startswith(_DRHALO_SIG):
        raise FileFormatError(
            'Not a Dr. Halo bitmap .FON: incorrect signature '
            f'{start[:len(_DRHALO_SIG)]}.'
        )
    width = int(le.int16.read_from(instream))
    height = int(le.int16.read_from(instream))
    if not height or not width:
        raise UnsupportedError(
            'Not a Dr. Halo bitmap .FON: may be stroked format.'
        )
    font = load_bitmap(
        instream, width=width, height=height, first_codepoint=first_codepoint,
    )
    font = font.modify(source_format='Dr. Halo')
    return font



"""
monobit.storage.wrapperformats.intelhex - binary files embedded in Intel Hex files

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import Path
from io import BytesIO

from monobit.base import FileFormatError
from ..streams import Stream, DelayedWriterStream
from ..base import encoders, decoders
from ...base.binary import ceildiv


@decoders.register(
    name='intel',
    patterns=(
        # '*.hex',
        '*.mcs', '*.int', '*.ihex', '*.ihe', '*.ihx'
    ),
    magic=(b':0',),
    # ends with b':00000001FF'
)
def decode_intel(instream):
    """Decode an intel hex-encoded stream."""
    infile = instream.text
    datadict = {}
    checksum = 0
    size = 0
    offset = 0
    for line in infile:
        _, _, dataline = line.partition(':')
        dataline = dataline.strip()
        if not dataline:
            continue
        try:
            databytes = int(dataline[:2], 16)
            address = offset + int.from_bytes(
                bytes.fromhex(dataline[2:6]),
                byteorder='big'
            )
            hexcode = dataline[6:8]
            checksum += int(dataline[-2:], 16)
            payload = bytes.fromhex(dataline[8:-2])
            if hexcode == '00':
                # data code
                if len(payload) != databytes:
                    logging.warning('Incorrect payload length.')
                datadict[address] = payload
                size = max(size, address+len(payload))
            elif hexcode == '01':
                # end of file code
                break
            elif hexcode == '02':
                # extended segment address
                offset = int.from_bytes(payload, byteorder='big') << 4
            elif hexcode == '04':
                # extended linear addresss
                offset = int.from_bytes(payload, byteorder='big') << 16
            elif hexcode in ('03', '05'):
                pass
            else:
                logging.warning(f'Ignoring unknown hex code {hexcode}.')
        except (IndexError, ValueError):
            raise FileFormatError('Malformed Intel Hex file.')
    data = bytearray(size)
    for address, payload in datadict.items():
        data[address:address+len(payload)] = payload
    name = Path(instream.name).stem
    return Stream.from_data(data, mode='r', name=name)


@encoders.register(linked=decode_intel)
def _encode_intel(outstream, *, chunk_size:int=32):
    """
    Decode an intel hex-encoded stream.

    chunk_size: byte count of a hex record line
    """
    encode_func = _do_encode_intel
    name = Path(outstream.name).stem
    return DelayedWriterStream(
        outstream, encode_func, name, chunk_size=chunk_size
    )


def _do_encode_intel(data, outstream, *, chunk_size):
    # split into groups of 32 bytes
    outfile = outstream.text
    # current extended linear address
    extended_address = -1
    offset = 0
    with BytesIO(data) as instream:
        while True:
            # in the last round this may be less than chunk_size long
            payload = instream.read(chunk_size)
            if not payload:
                # end of file marker
                outfile.write(':00000001FF\n')
                return
            sum_bytes = sum(payload)
            # skip over null chunks, except the last to ensure file length
            if not sum_bytes and len(payload) == chunk_size:
                continue
            offset_hi, offset_lo = divmod(offset, 0x10000)
            if offset_hi > extended_address:
                extended_address = offset_hi
                ea_checksum = (
                    0x100 - sum(offset_hi.to_bytes(2, byteorder='big'))
                    - 0x02 - 0x04
                ) % 0x100
                outfile.write(f':02000004{offset_hi:04X}{ea_checksum:02X}\n')
            checksum = (
                0x100
                - sum_bytes - len(payload)
                - sum(offset_lo.to_bytes(2, byteorder='big'))
            ) % 0x100
            outfile.write(
                f':{len(payload):02X}{offset_lo:04X}00'
                f'{payload.hex().upper()}{checksum:02X}\n'
            )
            offset += chunk_size



"""
monobit.core - tools for working with monochrome bitmap fonts

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .pack import Pack
from .font import Font, FontProperties, CUSTOM_NAMESPACE
from .glyph import Glyph, KernTable
from .raster import Raster
from .labels import Label, Char, Codepoint, Tag, strip_matching
from .vector import StrokePath, StrokeMove



"""
monobit.encoding - encoding classes

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from pathlib import Path

from .base import NotFoundError, EncodingName
from .registry import EncodingRegistry
from .charmaps import Charmap, Unicode, EncoderLoader
from .base import Encoder
from .indexers import Indexer
from .taggers import Tagmap
from .definitions import encodings
from ..core.labels import to_labels


def encoder(initialiser):
    """Retrieve or create a charmap from object or string."""
    if isinstance(initialiser, Encoder):
        return initialiser
    if initialiser is None or not str(initialiser):
        return None
    initialiser = str(initialiser)
    # numeric ranges - interpreted as indexer
    if initialiser[:1].isdigit():
        initialiser = to_labels(initialiser)
        return Indexer(code_range=initialiser)
    try:
        return encodings[initialiser]
    except KeyError:
        pass
    # if not registered, see if it is a file spec we can load
    filename, _, format = initialiser.partition(':')
    if not format:
        format = Path(filename).suffix[1:]
    try:
        if format == 'tbl':
            return Indexer.load(filename)
        return EncoderLoader(filename, format=format)()
    except (EnvironmentError, NotFoundError):
        return None



"""
monobit.storage.wrapperformats.compressors - single-file compression wrappers

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import gzip
import lzma
import bz2
from pathlib import Path

from monobit.base import FileFormatError
from ..streams import Stream
from ..base import encoders, decoders


###############################################################################
# gzip

@decoders.register(
    name='gzip',
    magic=(b'\x1f\x8b',),
    patterns=('*.gz',),
)
def decode_gzip(instream):
    """Decode a gzip-compressed stream."""
    stream = gzip.open(instream, mode='rb')
    try:
        stream.peek(0)
    except gzip.BadGzipFile as e:
        raise FileFormatError(e) from e
    name = Path(instream.name).stem
    return Stream(stream, mode='r', name=name)


@encoders.register(linked=decode_gzip)
def encode_gzip(outstream):
    """Encode to a gzip-compressed stream."""
    stream = gzip.open(outstream, mode='wb')
    name = Path(outstream.name).stem
    return Stream(stream, mode='w', name=name)


###############################################################################
# lzma

@decoders.register(
    name='xz',
    magic=(b'\xFD7zXZ\x00',),
    patterns=('*.xz',),
)
def decode_xz(instream):
    return _decode_lzma_or_xz(instream)


@decoders.register(
    name='lzma',
    # the magic is a 'maybe'
    magic=(b'\x5d\0\0',),
    patterns=('*.lzma',),
)
def decode_lzma(instream):
    return _decode_lzma_or_xz(instream)


def _decode_lzma_or_xz(instream):
    """Decode a lzma or xz-compressed stream."""
    stream = lzma.open(instream, mode='rb')
    try:
        stream.peek(0)
    except lzma.LZMAError as e:
        raise FileFormatError(e) from e
    name = Path(instream.name).stem
    return Stream(stream, mode='r', name=name)


@encoders.register(linked=decode_lzma)
def encode_lzma(outstream):
    """Encode to a lzma-compressed stream."""
    stream = lzma.open(outstream, mode='wb', format=lzma.FORMAT_ALONE)
    name = Path(outstream.name).stem
    return Stream(stream, mode='w', name=name)


@encoders.register(linked=decode_xz)
def encode_xz(outstream):
    """Encode to a xz-compressed stream."""
    stream = lzma.open(outstream, mode='wb', format=lzma.FORMAT_XZ)
    name = Path(outstream.name).stem
    return Stream(stream, mode='w', name=name)


###############################################################################
# bzip2

@decoders.register(
    name='bzip2',
    magic=(b'BZh',),
    patterns=('*.bz2',),
)
def decode_bzip2(instream):
    """Decode a bzip2-compressed stream."""
    stream = bz2.open(instream, mode='rb')
    try:
        stream.peek(0)
    except OSError as e:
        raise FileFormatError(e) from e
    name = Path(instream.name).stem
    return Stream(stream, mode='r', name=name)


@encoders.register(linked=decode_bzip2)
def encode_bzip2(outstream):
    """Encode to a bzip2-compressed stream."""
    stream = bz2.open(outstream, mode='wb')
    name = Path(outstream.name).stem
    return Stream(stream, mode='w', name=name)



"""
monobit.storage.fontformats.pcgeos - PC/GEOS 2.0+ (aka GeoWorks, NewDeal, BreadBox)

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate


from monobit.storage import loaders, savers
from monobit.core import Font, Glyph
from monobit.base import Props, reverse_dict, FileFormatError, UnsupportedError
from monobit.base.struct import StructError, bitfield, flag, little_endian as le
from monobit.base.binary import ceildiv

from monobit.storage.utils.limitations import make_contiguous


_BSWF_SIG = b'BSWF'


_FontFileInfo = le.Struct(
    # > "BSWF"
    FFI_signature='4s',
    # > minor version (0)
    FFI_minorVer='byte',
    # > major version (1)
    FFI_majorVer='byte',
    # > size of font info section
    # i.e. size of FontInfo + PointSizeTable, minus 1  (?)
    FFI_headerSize='word',
)

# typedef ByteFlags FontAttrs;
#define FA_USEFUL       0x80
#define FA_FIXED_WIDTH  0x40
#define FA_ORIENT       0x20
#define FA_OUTLINE      0x10
#define FA_FAMILY       0x0f
#define FA_FAMILY_OFFSET 0

_FontAttrs = le.Struct(
    FA_FAMILY=bitfield('uint8', 4),
    FA_OUTLINE=bitfield('uint8', 1),
    FA_ORIENT=bitfield('uint8', 1),
    FA_FIXED_WIDTH=bitfield('uint8', 1),
    FA_USEFUL=bitfield('uint8', 1),
)


# typedef byte FontFamily;
#define FF_NON_PORTABLE 0x0007
#define FF_SPECIAL 0x0006
#define FF_MONO 0x0005
#define FF_SYMBOL 0x0004
#define FF_ORNAMENT 0x0003
#define FF_SCRIPT 0x0002
#define FF_SANS_SERIF 0x0001
#define FF_SERIF 0x0000
_FAMILY_TO_STYLE = {
    0x0004: 'symbol',
    0x0002: 'script',
    0x0001: 'sans serif',
    0x0000: 'serif',
}
_STYLE_TO_FAMILY = reverse_dict(_FAMILY_TO_STYLE)


_FontInfo = le.Struct(
    # > font ID
    FI_fontID='word',
    # > font manufacturer ID
    # this is more of a "which vendor's format" field - for bitmap it is 0
    FI_maker='word',
    # > font family, etc.
    FI_family=_FontAttrs,
    # > typeface name
    FI_faceName='20s',
    # > start of table of point sizes
    FI_pointSizeTab='word',
    # > end of bitmap faces
    # actually, it's the end of the table of point sizes?
    FI_pointSizeEnd='word',
    # > start of outline table
    FI_outlineTab='word',
    # > end of outline faces
    FI_outlineEnd='word',
    # for "DBCS_PCGEOS" there are *two* byte fields here,
    #     FI_firstChar: char               ;first char included
    #     FI_lastChar: char               ;last char included
    # word-alignment
    unused='byte',
)


# typedef ByteFlags TextStyle;
#define TS_OUTLINE				0x40
#define TS_BOLD 				0x20
#define TS_ITALIC 				0x10
#define TS_SUPERSCRIPT 				0x08
#define TS_SUBSCRIPT 				0x04
#define TS_STRIKE_THRU 				0x02
#define TS_UNDERLINE 				0x01
_TextStyle = le.Struct(
    TS_UNDERLINE=flag,
    TS_STRIKE_THRU=flag,
    TS_SUBSCRIPT=flag,
    TS_SUPERSCRIPT=flag,
    TS_ITALIC=flag,
    TS_BOLD=flag,
    TS_OUTLINE=flag,
    unknown=flag,
)

def _style_to_decoration(style):
    return ' '.join(_deco for _deco in (
        'underline' if style.TS_UNDERLINE else '',
        'strikethrough' if style.TS_STRIKE_THRU else '',
        'subscript' if style.TS_SUBSCRIPT else '',
        'superscript' if style.TS_SUPERSCRIPT else '',
        'outline' if style.TS_OUTLINE else '',
    ) if _deco)


def _to_text_style(font):
    return _TextStyle(
        TS_UNDERLINE='underline' in font.decoration,
        TS_STRIKE_THRU='strikethrough' in font.decoration,
        TS_SUBSCRIPT='subscript' in font.decoration,
        TS_SUPERSCRIPT='superscript' in font.decoration,
        TS_ITALIC=font.slant == 'italic',
        TS_BOLD=font.weight == 'bold',
        TS_OUTLINE='outline' in font.decoration,
    )


_PointSizeEntry = le.Struct(
    # > style of face
    PSE_style=_TextStyle,
    # note that there are different versions of this struct
    # the DBCS_PCGEOS version has this, instead of below WBFixed field
        # PSE_pointSize       sbyte                   ;integer point size <128
        # PSE_charSet         FontCharSet             ;characters in section
    # > point size
    # > PSE_pointSize       WBFixed                 ;point size
    # "WBFixed" format - assume byte fractional part, byte integer part
    # though elsewhere I seem to need byte + word for WBFixed??
    PSE_pointSize_int='int8',
    PSE_pointSize_frac='uint8',
    # > size of data
    PSE_dataSize='word',
    # > position in file
    # PSE_dataPosLo='word',
    # > position in file
    # PSE_dataPosHi='word',
    PSE_dataPos='dword',
    # word-alignment
    unused='byte',
)

_WBFixed = le.Struct(
    frac='byte',
    int='word',
)

def _wbfixed_to_float(wbfixed_value):
    """Convert WBFixed value to float."""
    return wbfixed_value.int + wbfixed_value.frac / 256


def _float_to_wbfixed(float_value):
    """Convert float value to WBFixed."""
    return _WBFixed(
        frac=int((float_value-int(float_value)) * 256),
        int=int(float_value),
    )


# typedef ByteFlags FontBufFlags;
#define FBF_DEFAULT_FONT    0x80
#define FBF_MAPPED_FONT     0x40
#define FBF_IS_OUTLINE      0x10
#define FBF_IS_REGION       0x08
#define FBF_IS_COMPLEX      0x04
#define FBF_IS_INVALID      0x02
_FontBufFlags = le.Struct(
    unknown_0x01=flag,
    FBF_IS_INVALID=flag,
    FBF_IS_COMPLEX=flag,
    FBF_IS_REGION=flag,
    FBF_IS_OUTLINE=flag,
    unknown_0x20=flag,
    FBF_MAPPED_FONT=flag,
    FBF_DEFAULT_FONT=flag,
)


# typedef word FontMaker;
#define FM_PRINTER                              0xf000
#define FM_MICROLOGIC                           0xe000
#define FM_ATECH                                0xd000
#define FM_PUBLIC                               0xc000
#define FM_AGFA                                 0x4000
#define FM_BITSTREAM                            0x3000
#define FM_ADOBE                                0x2000
#define FM_NIMBUSQ                              0x1000
#define FM_BITMAP                               0x0000


_FontBuf = le.Struct(
    # > actual size of data (bytes)
    FB_dataSize='word',
    # > manufacturer ID
    # > FontMaker type
    FB_maker='word',
    # > average character width
    FB_avgwidth=_WBFixed,
    # > width of widest character
    FB_maxwidth=_WBFixed,
    # > offset to top of font box
    FB_heightAdjust=_WBFixed,
    # > height of characters
    FB_height=_WBFixed,
    # > height of accent portion.
    FB_accent=_WBFixed,
    # > top of lower case character boxes.
    FB_mean=_WBFixed,
    # > offset to top of ascent
    FB_baseAdjust=_WBFixed,
    # > position of baseline from top of font
    FB_baselinePos=_WBFixed,
    # > maximum descent (from baseline)
    FB_descent=_WBFixed,
    # > recommended external leading
    FB_extLeading=_WBFixed,
    # >   line spacing = FB_height +
    # >                  FB_extLeading +
    # >                  FB_heightAdjust
    # > number of kerning pairs
    FB_kernCount='word',
    # > offset to kerning pair table
    # > nptr.KernPair
    FB_kernPairPtr='word',
    # > offset to kerning value table
    # > nptr.BBFixed
    FB_kernValuePtr='word',
# if DBCS_PCGEOS
#     FB_firstChar        Chars           ; first char in section
#     FB_lastChar         Chars           ; last char in section
#     FB_defaultChar      Chars           ; default character
# else
    # > first char defined
    FB_firstChar='byte',
    # > last char defined
    FB_lastChar='byte',
    # > default character
    FB_defaultChar='byte',
# endif
    # > underline position (from baseline)
    FB_underPos=_WBFixed,
    # > underline thickness
    FB_underThickness=_WBFixed,
    # > position of the strike-thru
    FB_strikePos=_WBFixed,
    # > maximum above font box
    FB_aboveBox=_WBFixed,
    # > maximum below font box
    FB_belowBox=_WBFixed,
    # > Bounds are signed integers, in device coords, and are
    # > measured from the upper left of the font box where
    # > character drawing starts from.
    #
    # > minimum left side bearing
    FB_minLSB='int16',
    # ; minimum top side bound
    FB_minTSB='int16',
# if not DBCS_PCGEOS
    # > maximum bottom side bound
    FB_maxBSB='int16',
    # ; maximum right side bound
    FB_maxRSB='int16',
# endif
    # > height of font (invalid for rotation)
    FB_pixHeight='word',
    # > special flags
    FB_flags=_FontBufFlags,
    # > usage counter for this font
    FB_heapCount='word',
    # FB_charTable        CharTableEntry <>
)


_CharTableFlags = le.Struct(
    # > pad to a BYTE only
    # . TRUE if negative left-side bearing
    CTF_NEGATIVE_LSB=flag,
    # > TRUE if very tall
    CTF_ABOVE_ASCENT=flag,
    # > TRUE if very low
    CTF_BELOW_DESCENT=flag,
    # > TRUE if no data
    CTF_NO_DATA=flag,
    # > TRUE if first of a kern pair
    CTF_IS_FIRST_KERN=flag,
    # > TRUE if second of a kern pair
    CTF_IS_SECOND_KERN=flag,
    # > TRUE if character normally invisible
    CTF_NOT_VISIBLE=flag,
)

_CharTableEntry = le.Struct(
    # > Offset to data
    # > nptr.CharData
    CTE_dataOffset='word',
    # > character width
    CTE_width=_WBFixed,
    # > flags
    CTE_flags=_CharTableFlags,
    # > LRU count
    CTE_usage='word',
)


_CharData = le.Struct(
    # > width of picture data in bits
    CD_pictureWidth='byte',
    # > number of rows of data
    CD_numRows='byte',
    # > (signed) offset to first row
    CD_yoff='int8',
    # > (signed) offset to first column
    CD_xoff='int8',
    # > data for character
    # CD_data
)


@loaders.register(
    name='pcgeos',
    magic=(_BSWF_SIG,),
    patterns=('*.fnt',),
)
def load_pcgeos(instream):
    """Load a PC/GEOS v2.0+ bitmap font file."""
    font_file_info = _FontFileInfo.read_from(instream)
    logging.debug('FontFileInfo: %s', font_file_info)
    if font_file_info.FFI_signature != _BSWF_SIG:
        if font_file_info.FFI_signature[:2] == b'\xEF\xBE':
            raise UnsupportedError('PC/GEOS v1.2 fonts not supported.')
        if font_file_info.FFI_signature[:2] == b'\x80\x10':
            raise UnsupportedError('PC/GEOS v1.0 fonts not supported.')
        raise FileFormatError('Not a PC-GEOS font file: incorrect signature')
    font_info = _FontInfo.read_from(instream)
    logging.debug('FontInfo: %s', font_info)
    if font_info.FI_maker != 0:
        logging.warning(
            'Non-bitmap PC/GEOS font file; '
            'will try to extract bitmap glyphs, if any.'
        )
    # the +7 offset is strange, as the FFI header is 8 bytes long
    instream.seek(font_info.FI_pointSizeTab + 7)
    n_entries = (font_info.FI_pointSizeEnd - font_info.FI_pointSizeTab) //_PointSizeEntry.size
    point_size_table = (_PointSizeEntry * n_entries).read_from(instream)
    # the first FontBuf actually starts 1 byte earlier than where we are now
    # i.e. it overwrites the unused padding byte in the last entry
    fonts = []
    for point_size_entry in point_size_table:
        logging.debug('PointSizeEntry: %s', point_size_entry)
        instream.seek(point_size_entry.PSE_dataPos)
        font_buf = _FontBuf.read_from(instream)
        logging.debug('FontBuf: %s', font_buf)
        n_chars = font_buf.FB_lastChar - font_buf.FB_firstChar + 1
        char_table = (_CharTableEntry * n_chars).read_from(instream)
        glyphs = []
        for cp, char_table_entry in enumerate(
                char_table, font_buf.FB_firstChar
            ):
            # this appears to be necessary
            if char_table_entry.CTE_dataOffset == 0:
                continue
            # this is usually (?) a no-op at this point
            instream.seek(
                point_size_entry.PSE_dataPos + char_table_entry.CTE_dataOffset
            )
            try:
                char_data = _CharData.read_from(instream)
            except StructError:
                if not glyph:
                    raise
                break
            if char_table_entry.CTE_flags.CTF_NO_DATA:
                charbytes = b''
            else:
                byte_width = ceildiv(char_data.CD_pictureWidth, 8)
                byte_size = char_data.CD_numRows * byte_width
                # blank glyph still has one null row
                byte_size = max(1, byte_size)
                charbytes = instream.read(byte_size)
            glyph = Glyph.from_bytes(
                charbytes, width=char_data.CD_pictureWidth,
                shift_up=(
                    _wbfixed_to_float(font_buf.FB_baselinePos)
                    - (char_data.CD_numRows+char_data.CD_yoff)
                ),
                left_bearing=char_data.CD_xoff,
                right_bearing=(
                    _wbfixed_to_float(char_table_entry.CTE_width)
                    - char_data.CD_pictureWidth
                    - char_data.CD_xoff
                ),
                codepoint=cp,
            )
            glyphs.append(glyph)
        font = Font(
            glyphs,
            family=font_info.FI_faceName.decode('ascii', 'replace'),
            font_id=font_info.FI_fontID,
            point_size=(
                point_size_entry.PSE_pointSize_int +
                point_size_entry.PSE_pointSize_frac / 256
            ),
            average_width=_wbfixed_to_float(font_buf.FB_avgwidth),
            max_width=_wbfixed_to_float(font_buf.FB_maxwidth),
            # FB_heightAdjust=font_buf.FB_heightAdjust,
            x_height=_wbfixed_to_float(font_buf.FB_mean),
            # FB_baseAdjust=font_buf.FB_baseAdjust,
            ascent=(
                _wbfixed_to_float(font_buf.FB_height)
                - _wbfixed_to_float(font_buf.FB_accent)
                - _wbfixed_to_float(font_buf.FB_descent)
            ),
            descent=_wbfixed_to_float(font_buf.FB_descent),
            line_height=(
                _wbfixed_to_float(font_buf.FB_height)
                + _wbfixed_to_float(font_buf.FB_extLeading)
                + _wbfixed_to_float(font_buf.FB_heightAdjust)
            ),
            default_char=font_buf.FB_defaultChar,
            underline_descent=(
                _wbfixed_to_float(font_buf.FB_underPos) - _wbfixed_to_float(font_buf.FB_baselinePos)
            ),
            underline_thickness=_wbfixed_to_float(font_buf.FB_underThickness),
            strikethrough_ascent=(
                _wbfixed_to_float(font_buf.FB_baselinePos)
                - _wbfixed_to_float(font_buf.FB_strikePos) + 1
            ),
            decoration=_style_to_decoration(point_size_entry.PSE_style),
            slant='italic' if point_size_entry.PSE_style.TS_ITALIC else None,
            weight='bold' if point_size_entry.PSE_style.TS_BOLD else None,
            style=_FAMILY_TO_STYLE.get(font_info.FI_family.FA_FAMILY, None),
        )
        font = font.label(char_from='pc-geos')
        fonts.append(font)
    return fonts


@savers.register(linked=load_pcgeos)
def save_pcgeos(fonts, outstream):
    """Save to a PC/GEOS v2.0+ bitmap font file."""
    fonts, common_props = _prepare_pcgeos(fonts)
    fonts = tuple(
        _f.label(codepoint_from=_f.encoding).subset(codepoints=range(256))
        for _f in fonts if _f.glyphs
    )
    n_sizes = len(fonts)
    # FontFileInfo header
    font_file_info = _FontFileInfo(
        FFI_signature=_BSWF_SIG,
        FFI_minorVer=0,
        FFI_majorVer=1,
        FFI_headerSize=_FontInfo.size - 1 + _PointSizeEntry.size * n_sizes,
    )
    # strange 7 byte offset
    point_size_tab_offset = _FontFileInfo.size + _FontInfo.size - 7
    # FontInfo header
    font_info = _FontInfo(
        FI_fontID=int(common_props.font_id),
        FI_maker=0,
        FI_family=_FontAttrs(
            FA_FAMILY=(
                5 if common_props.fixed_width
                else _STYLE_TO_FAMILY.get(common_props.style, 0)
            ),
            FA_FIXED_WIDTH=common_props.fixed_width,
            FA_USEFUL=1,
        ),
        FI_faceName=common_props.family,
        FI_pointSizeTab=point_size_tab_offset,
        FI_pointSizeEnd=point_size_tab_offset + _PointSizeEntry.size * n_sizes,
    )
    # per-font sections
    data_offset = 0
    font_data = []
    point_size_table = []
    for font in fonts:
        data = _create_pcgeos_font_section(font)
        font_data.append(data)
        point_size_table.append(_PointSizeEntry(
            PSE_style=_to_text_style(font),
            PSE_pointSize_int=int(font.point_size),
            PSE_pointSize_frac=int(256*(font.point_size-int(font.point_size))),
            PSE_dataSize=len(data),
            PSE_dataPos=data_offset + (
                _FontFileInfo.size + _FontInfo.size
                # -1 as we will clip off the last byte
                + (_PointSizeEntry * n_sizes).size - 1
            ),
        ))
        data_offset += len(data)
    point_size_table = (_PointSizeEntry * n_sizes)(*point_size_table)
    # write out pcgeos file
    outstream.write(bytes(font_file_info))
    outstream.write(bytes(font_info))
    # note that we clip off the final null byte
    outstream.write(bytes(point_size_table)[:-1])
    for data in font_data:
        outstream.write(data)


def _create_pcgeos_font_section(font):
    """Create pcgeos font section for one font."""
    baseline = font.raster_size.y + min(_g.shift_up for _g in font.glyphs)
    first_char = int(min(font.get_codepoints()))
    last_char = int(max(font.get_codepoints()))
    glyphs = tuple(
        font.get_glyph(_cp, missing=None)
        for _cp in range(first_char, last_char + 1)
    )
    n_chars = len(glyphs)
    glyphbytes = tuple(
        (
            bytes(_CharData(
                CD_pictureWidth=_g.width,
                CD_numRows=_g.height,
                CD_yoff=(baseline - _g.height - _g.shift_up),
                CD_xoff=_g.left_bearing,
            ))
            # empty glyph still stores one byte of zeros
            + (_g.as_bytes() or b'\0')
        ) if _g is not None else (
            # missing glyph omitted altogether, but needs a record in CharTable
            b''
        )
        for _g in glyphs
    )
    glyphdata = b''.join(glyphbytes)
    glyph_offsets = accumulate(
        (len(_b) for _b in glyphbytes[:-1]),
        initial=0
    )
    header_size = _FontBuf.size + (_CharTableEntry * n_chars).size
    font_buf = _FontBuf(
        FB_dataSize=(
            len(glyphdata) + _FontBuf.size
            + (_CharTableEntry * n_chars).size
        ),
        FB_maker=0,
        FB_avgwidth=_float_to_wbfixed(font.average_width),
        FB_maxwidth=_float_to_wbfixed(font.max_width),
        # > offset to top of font box
        # FB_heightAdjust=_WBFixed,
        FB_height=_float_to_wbfixed(font.raster_size.y),
        FB_accent=_float_to_wbfixed(font.raster_size.y - font.pixel_size),
        # > top of lower case character boxes.
        FB_mean=_float_to_wbfixed(font.x_height),
        # > offset to top of ascent
        # FB_baseAdjust=_WBFixed,
        FB_baselinePos=_float_to_wbfixed(baseline),
        FB_descent=_float_to_wbfixed(font.descent),
        FB_extLeading=_float_to_wbfixed(font.line_height - font.raster_size.y),
        # I don't know how kerning works in this format and have no samples
        FB_kernCount=0,
        # pointer values as in Berkeley.fnt
        FB_kernPairPtr=header_size,
        FB_kernValuePtr=header_size,
        FB_firstChar=first_char,
        FB_lastChar=last_char,
        FB_defaultChar=int(font.get_default_glyph().codepoint or first_char),
        FB_underPos=_float_to_wbfixed(baseline + font.underline_descent),
        FB_underThickness=_float_to_wbfixed(font.underline_thickness),
        # > position of the strike-thru
        FB_strikePos=_float_to_wbfixed(baseline - font.strikethrough_ascent + 1),
        # > maximum above font box
        # FB_aboveBox=_WBFixed,
        # > maximum below font box
        # FB_belowBox=_WBFixed,
        FB_minLSB=min(_g.left_bearing for _g in font.glyphs),
        FB_minTSB=0,
        FB_maxBSB=font.raster_size.y,
        FB_maxRSB=max(_g.advance_width for _g in font.glyphs),
        FB_pixHeight=font.raster_size.y,
        # # > special flags
        # FB_flags=_FontBufFlags,
        # this is the value used in font samples - perhaps a sentinel? 666 dec
        FB_heapCount=0x29a,
    )
    char_table = (_CharTableEntry * n_chars)(*(
        _CharTableEntry(
            CTE_dataOffset=0,
            CTE_flags=_CharTableFlags(
                CTF_NO_DATA=True,
            )
        ) if _glyph is None else
        _CharTableEntry(
            CTE_dataOffset=_FontBuf.size + (_CharTableEntry * n_chars).size + _ofs,
            CTE_width=_float_to_wbfixed(_glyph.advance_width),
            CTE_flags=_CharTableFlags(
                CTF_NEGATIVE_LSB=(_glyph.left_bearing < 0),
                # > TRUE if very tall
                # CTF_ABOVE_ASCENT=flag,
                # > TRUE if very low
                # CTF_BELOW_DESCENT=flag,
                # if we set this, omit the 1-byte placeholder for glyph data
                # CTF_NO_DATA=flag,
                CTF_IS_FIRST_KERN=False,
                CTF_IS_SECOND_KERN=False,
                # CTF_NOT_VISIBLE=flag,
            ),
            CTE_usage=0,
        )
        for _glyph, _ofs in zip(glyphs, glyph_offsets)
    ))
    return b''.join((
        bytes(font_buf),
        bytes(char_table),
        glyphdata,
    ))


def _prepare_pcgeos(fonts):
    """Validate fonts for storing in PC-GEOS format; extract metadata."""
    if len(set(_f.family for _f in fonts)) > 1:
        raise UnsupportedError(
            'PC-GEOS font file can only store fonts from one family.'
        )
    common_props = _get_metadata(fonts[0])
    return fonts, common_props


def _get_metadata(font):
    try:
        font_id = int(font.font_id)
    except ValueError:
        font_id = 0
    return Props(
        font_id=font_id,
        family=font.family.encode('ascii', 'replace'),
        fixed_width=font.spacing in ('character-cell', 'monospace'),
        style=font.style,
    )



"""
monobit.storage.fontformats.text.consoleet - consoleet / vfontas / hxtools format

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from pathlib import Path
import logging

from monobit.storage.base import container_loaders, container_savers
from monobit.core import Font, Glyph

from monobit.storage.utils.limitations import ensure_single
from monobit.storage.utils.perglyph import loop_load, loop_save


@container_loaders.register(name='consoleet')
def load_clt(location):
    """Load font from consoleet files."""
    # this format consists of separate image files, without a manifest
    glyphs = loop_load(location, _read_clt_glyph)
    return Font(glyphs, source_name=location.path)


@container_savers.register(linked=load_clt)
def save_clt(fonts, location):
    """
    Save font to consoleet files.
    """
    loop_save(
        fonts, location,
        prefix='', suffix='txt', save_func=_write_clt_glyph,
    )


def _read_clt_glyph(instream):
    text = instream.text
    name = instream.name
    codepoint = Path(name).stem
    magic = text.readline().strip()
    if magic != 'PCLT':
        return Glyph()
    width, _, height = text.readline().strip().partition(' ')
    glyphtext = tuple(text.read().splitlines())
    return Glyph(
        glyphtext,
        inklevels='.#',
        # encoding is not specified by spec or file - can be unicode or codepage
        codepoint=f'0x{codepoint}',
    ).shrink(factor_x=2)


def _write_clt_glyph(glyph, outstream):
    text = outstream.text
    text.write('PCLT\n')
    text.write(f'{glyph.width} {glyph.height}\n')
    text.write(glyph.as_text(inklevels=('..', '##'), end='\n'))



"""
monobit.storage.containerformats.binhex - BinHex containers

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from binascii import crc_hqx
from itertools import zip_longest
from functools import partial

from monobit.base.struct import big_endian as be, little_endian as le
from monobit.base import FileFormatError
from ..streams import Stream
from ..magic import Magic, Sentinel
from ..base import containers
from ..containers import FlatFilterContainer


###############################################################################
# BinHex 4.0 container
# https://files.stairways.com/other/binhex-40-specs-info.txt

_BINHEX_CODES = (
    '''!"#$%&'()*+,-012345689@ABCDEFGHIJKLMNPQRSTUVXYZ[`abcdefhijklmpqr'''
)
_BINHEX_CODEDICT = {_BINHEX_CODES[_i]: _i for _i in range(64)}

_BINHEX_HEADER = be.Struct(
    type='4s',
    auth='4s',
    flag='uint16',
    dlen='uint32',
    rlen='uint32',
)
_CRC = be.Struct(
    crc='uint16',
)

def sixbitdecode(encoded, codedict, byteswap=False):
    """Deocde six-bit ascii representation into bytes."""
    # decode into 6-bit ints
    try:
        data = (codedict[_c] for _c in encoded)
    except KeyError as e:
        raise FileFormatError(f"Unexpected character '{e}' in encoding.") from e
    if byteswap:
        # reverse in groups of 4 code bytes, for binscii
        args = [iter(data)] * 4
        data = b''.join(bytearray(reversed(_chunk)) for _chunk in zip(*args))
    # convert to bit sequence
    bits = ''.join(bin(_d)[2:].zfill(6) for _d in data)
    # group into chunks of 8
    args = [iter(bits)] * 8
    octets = (''.join(_t) for _t in zip_longest(*args, fillvalue='0'))
    # convert to bytes
    bytestr = bytes(int(_s, 2) for _s in octets)
    return bytestr


@containers.register(
    name='binhex',
    magic=(
        b'(This file must be converted',
        b'\r(This file must be converted',
    ),
    patterns=('*.hqx',),
)
class BinHex(FlatFilterContainer):
    """BinHex 4.0 loader."""

    def decode(self, name):
        """
        Decode data and resource fork from BinHex container.
        """
        return super().decode(name)

    def encode(self, name):
        """
        Writing to BinHex is not supported.
        """
        raise ValueError(
            'Writing to BinHex is not supported.'
        )

    def decode_all(self, stream):
        """Parse a BinHex 4.0 file."""
        front, binhex, *back = stream.text.read().split(':')
        if 'BinHex 4.0' not in front:
            logging.warning('No BinHex 4.0 signature found.')
        back = ''.join(back).strip()
        if back:
            logging.warning('Additional data found after BinHex section: %r', back)
        binhex = ''.join(binhex.split('\n'))
        bytestr = sixbitdecode(binhex, _BINHEX_CODEDICT)
        # find run-length encoding marker
        chunks = bytestr.split(b'\x90')
        out = bytearray(chunks[0])
        for c in chunks[1:]:
            if c:
                # run-length byte
                repeat = c[0]
            else:
                # ...\x90\x90... -> ...', '', '...
                repeat = 0x90
            if not repeat:
                # zero-byte is placeholder for just 0x90
                out += b'\x90'
            else:
                # apply RLE. the last byte counts as the first of the run
                out += out[-1:] * (repeat-1)
            out += c[1:]
        # decode header
        length = out[0]
        name = bytes(out[1:1+length]).decode('mac-roman')
        if out[1+length] != 0:
            logging.warning('No null byte after name')
        header = _BINHEX_HEADER.from_bytes(out, 2+length)
        logging.debug(header)
        offset = 2 + length + _BINHEX_HEADER.size
        crc_header = out[:offset]
        hc = _CRC.from_bytes(out, offset)
        offset += _CRC.size
        if crc_hqx(crc_header, 0) != hc.crc:
            logging.error('CRC fault in header')
        data = out[offset:offset+header.dlen]
        offset += header.dlen
        dc = _CRC.from_bytes(out, offset)
        offset += _CRC.size
        rsrc = out[offset:offset+header.rlen]
        offset += header.rlen
        rc = _CRC.from_bytes(out, offset)
        if crc_hqx(data, 0) != dc.crc:
            logging.error('CRC fault in data fork')
        if crc_hqx(rsrc, 0) != rc.crc:
            logging.error('CRC fault in resource fork')
        if name:
            return {
                f'data/{name}': data,
                f'rsrc/{name}': rsrc,
            }
        else:
            return {
                f'data': data,
                f'rsrc': rsrc,
            }


###############################################################################
# BinSCII
# based on https://mirrors.apple2.org.za/ground.icaen.uiowa.edu/Mirrors/uni-kl/doc/binscii.format


_BINSCII_HEADER = le.Struct(
    # size fields are little-endian uint24, not supported by Struct
    filesize=be.uint8 * 3,
    segstart=be.uint8 * 3,
    acmode='byte',
    filetype='byte',
    auxtype='word',
    storetype='byte',
    blksize='word',
    credata='word',
    cretime='word',
    moddate='word',
    modtime='word',
    seglen=be.uint8 * 3,
    crc='word',
    filler='byte',
)

_BINSCII_SENTINEL = 'FiLeStArTfIlEsTaRt'


@containers.register(
    name='binscii',
    magic=(
        # this can occur on any line, probably not the first...
        Sentinel(_BINSCII_SENTINEL.encode('ascii')),
    ),
    patterns=('*.bsc', '*.bsq', '*.bns'),
)
class BinSCII(FlatFilterContainer):
    """BinSCII loader."""

    def decode(self, name):
        """
        Decode file from BinSCII container.
        """
        return super().decode(name)

    def encode(self, name):
        """
        Writing to BinSCII is not supported.
        """
        raise ValueError(
            'Writing to MacBinary is not supported.'
        )

    def decode_all(self, instream):
        """Parse a BinSCII file."""
        instream = instream.text
        output = {}
        while True:
            # locate binscii section sentinel
            for line in instream:
                if line.strip() == _BINSCII_SENTINEL:
                    break
            else:
                break
            # encoding alphabet
            encoding = instream.readline().strip()
            codedict = {_c: _i for _i, _c in enumerate(encoding)}
            # name and header
            metadata = instream.readline().strip()
            name_len = encoding.index(metadata[0]) + 1
            name = metadata[1:16]
            name = name[:name_len]
            binsciidecode = partial(sixbitdecode, codedict=codedict, byteswap=True)
            headerbytes = binsciidecode(metadata[16:16+36])
            header = _BINSCII_HEADER.from_bytes(headerbytes)
            # size fields are little-endian uint24, not supported by Struct
            filesize = int.from_bytes(header.filesize, 'little')
            segstart = int.from_bytes(header.segstart, 'little')
            seglen = int.from_bytes(header.seglen, 'little')
            logging.debug('binscii header: %s', header)
            logging.debug('filesize=%d segstart=%d seglen=%d', filesize, segstart, seglen)
            header_crc = crc_hqx(headerbytes[:-3], 0)
            if header_crc != header.crc:
                logging.warning(f"CRC failure while decoding '{name}'")
            # data section
            decoded = []
            crcline = ''
            for dataline in instream:
                dataline = dataline.strip()
                if len(dataline) != 64:
                    crcline = dataline
                    break
                decoded.append(binsciidecode(dataline))
            data = b''.join(decoded)
            crc_bytes = binsciidecode(crcline)
            crc_target = int.from_bytes(crc_bytes[:2], byteorder='little')
            if crc_hqx(data, 0) != crc_target:
                logging.warning(f"CRC failure while decoding '{name}'")
            if name not in output:
                output[name] = bytearray(filesize)
            output[name][segstart:segstart+seglen] = data
        if not output:
            raise FileFormatError(
                f"No BinSCII section found in file '{instream.name}'."
            )
        return output



"""
monobit.storage.fontformats.geos - C64 GEOS font files

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from datetime import datetime
from itertools import count, accumulate

from monobit.storage import loaders, savers, Stream, Magic
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Font, Glyph, Raster
from monobit.base.struct import little_endian as le
from monobit.base.binary import ceildiv

from monobit.storage.utils.limitations import ensure_single, make_contiguous


###############################################################################
# Font record in GEOS VLIR file
# https://www.lyonlabs.org/commodore/onrequest/geos/geos-fonts.html

_HEADER = le.Struct(
    baseline='uint8',
    stride='uint16',
    height='uint8',
    index_offset='uint16',
    bitstream_offset='uint16',
)

# characters 0x20 - 0x7f
# however, for sample files, the value there doesn't apper to be a valid offset
# nor is there a coherent image in the strike past offset[95]
_OFFSETS = le.uint16.array(97)


@loaders.register(
    name='vlir',
)
def load_geos_vlir_record(instream, *, extract_del:bool=False):
    """
    Load a bare GEOS font VLIR.

    extract_del: extract the unused DEL glyph 0x7f (defauult: False)
    """
    header = _HEADER.read_from(instream)
    logging.debug('header: %s', header)
    instream.seek(header.index_offset)
    offsets = _OFFSETS.read_from(instream)
    instream.seek(header.bitstream_offset)
    strikebytes = instream.read(header.height * header.stride)
    strikebytes = strikebytes.ljust(header.height * header.stride, b'\0')
    strike = Raster.from_bytes(
        strikebytes,
        header.stride * 8, header.height,
    )
    # clip out glyphs
    glyphs = tuple(
        Glyph(
            strike.crop(left=_offset, right=max(0, header.stride*8 - _next)),
            codepoint=_cp,
            shift_up=-(header.height-header.baseline-1)
        )
        for _offset, _next, _cp in zip(offsets, offsets[1:], count(0x20))
    )
    # drop glyph 0x7f (DEL) as it's never used and often garbage
    if not extract_del:
        glyphs = glyphs[:-1]
    props = dict(
        descent=header.height-header.baseline-1,
        ascent=header.baseline+1,
        encoding='ascii',
    )
    return Font(glyphs, **props).label()


@savers.register(linked=load_geos_vlir_record)
def save_geos_vlir_record(fonts, outstream):
    """Save font to a bare GEOS font VLIR."""
    font = ensure_single(fonts)
    font = _prepare_font_for_geos_vlir_record(font)
    outstream.write(_create_geos_vlir_record(font))


def _prepare_font_for_geos_vlir_record(font):
    """Prepare a font to be stored in a geos VLIR record."""
    font = font.label(match_graphical=False, match_whitespace=False)
    font = font.subset(chars=(chr(_c) for _c in _GEOS_RANGE))
    font = font.label(
        codepoint_from='ascii', overwrite=True,
        match_graphical=False, match_whitespace=False,
    )
    font = make_contiguous(font, full_range=_GEOS_RANGE, missing='empty')
    font = font.equalise_horizontal()
    return font


def _create_geos_vlir_record(font):
    """Save font to a bare GEOS font VLIR record."""
    # generate strike
    offsets = (0, *accumulate(_g.width for _g in font.glyphs))
    stride = ceildiv(offsets[-1], 8)
    rasters = [_g.pixels for _g in font.glyphs]
    # glyphs have been equalised so we can refer to the first one
    height = rasters[0].height
    baseline = height + font.glyphs[0].shift_up - 1
    rasters.append(
        Raster.blank(width=stride*8 - offsets[-1], height=height)
    )
    strike = Raster.concatenate(*rasters)
    offset_table = _OFFSETS(*offsets)
    header = _HEADER(
        baseline=baseline,
        stride=stride,
        height=height,
        index_offset=_HEADER.size,
        bitstream_offset=_HEADER.size + _OFFSETS.size,
    )
    logging.debug('header: %s', header)
    return b''.join((
        bytes(header),
        bytes(offset_table),
        strike.as_bytes(),
    ))


###############################################################################
# GEOS VLIR file in CONVERT (CVT) container
# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT

# the CVT contains a signature block, info block and record block, followed by
# one or more VLIRs (whose starting sectors are given by the record block)
# each such record is one sectore, i.e. 256 bytes.
# However in the CVT files the two initial bytes, a linked-list pointer,
# are left out.

# https://ist.uwaterloo.ca/~schepers/formats/D64.TXT
# > File Type
_C64_FILETYPES = {
    0x00: 'Scratched (deleted file entry)',
    0x80: 'DEL',
    0x81: 'SEQ',
    0x82: 'PRG',
    0x83: 'USR',
    0x84: 'REL',
}

# > GEOS file structure
_GEOS_STRUCTURES = {
    0x00: 'Sequential',
    0x01: 'VLIR file',
}
# > GEOS filetype
_GEOS_FILETYPES = {
    0x00: 'Non-GEOS (normal C64 file)',
    0x01: 'BASIC',
    0x02: 'Assembler',
    0x03: 'Data file',
    0x04: 'System File',
    0x05: 'Desk Accessory',
    0x06: 'Application',
    0x07: 'Application Data (user-created documents)',
    0x08: 'Font File',
    0x09: 'Printer Driver',
    0x0A: 'Input Driver',
    0x0B: 'Disk Driver (or Disk Device)',
    0x0C: 'System Boot File',
    0x0D: 'Temporary',
    0x0E: 'Auto-Execute File',
    # >  0F-FF - Undefined
}

_GEOS_FONT_TYPE = 0x08

# 30 bytes 0x1e
# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
_DIR_BLOCK = le.Struct(
    # > C64 filetype (see the section on D64 for an explanation)
    # > REL files are not allowed.
    filetype='uint8',
    # > Starting track/sector (02/02 from above) of C64 file if GEOS
    # > filetype is $00. If GEOS filetype is non-zero,  track/sector
    # > of single-sector RECORD block
    sector='uint16',
    # > Filename (in ASCII, padded with $A0, case varies)
    filename='16s',
    # > Track/sector location of info block
    info_sector='uint16',
    # > GEOS file structure
    geos_structure='uint8',
    # > GEOS filetype
    geos_filetype='uint8',
    # > Year (1900 + value)
    year='uint8',
    # > Month (1-12, $01 to $0C)
    month='uint8',
    # > Day (1-31, $01 to $1F)
    day='uint8',
    # > Hour (0-23, $00 to $17) in military format
    hour='uint8',
    # > Minute (0-59, $00 to $3B)
    minute='uint8',
    # > Filesize, in sectors (low/high byte order)
    filesize='uint16',
)

# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT
_SIG_BLOCK = le.Struct(
    # 0x1e   30
    # b'PRG formatted GEOS file V1.0'
    # b'SEQ formatted GEOS file V1.0'
    signature='28s',
    # 0x3a   58
    notes=le.uint8.array(196),
    # 0xfe  254
)

# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
_INFO_BLOCK = le.Struct(
    # 0x02 / 0xfe - cvt leaves out the word-size pointer at the start
    #        so sectors are only 254 bytes long
    # > Information sector ID bytes (03 15 BF). The "03" is  likely
    # > the bitmap width, and the "15" is likely the bitmap height,
    # > but rare exceptions do exist to this!
    id_bytes='3s', # > 03 15 bf
    # 0x101
    # > Icon bitmap (sprite format, 63 bytes)
    icon=le.uint8.array(63),
    # 0x140
    # > C64 filetype (same as that from the directory entry)
    filetype='uint8',
    # > GEOS filetype (same as that from the directory entry)
    geos_filetype='uint8',
    # > GEOS file structure (same as that from the dir entry)
    geos_structure='uint8',
    # 0x143
    # > Program load address
    load_address='uint16',
    # > Program end address (only with accessories)
    end_address='uint16',
    # > Program start address
    start_address='uint16',
    # 0x149
    # > Class text (terminated with a $00)
    class_text='20s',

    # 0x61 / 0x15d
    # > Author (with application data: name  of  application  disk,
    # > terminated with a $00. This string may not  necessarily  be
    # > set, or it may contain invalid data)
    #author='20s',
    # 0x75 / 0x171
    # > 75-88: If a document, the name of the application that created it.
    ##application='20s',
    # 0x89/0x185
    # > 89-9F: Available for applications, unreserved.
    #unreserved='23s',

    # here the font INFO section diverges from the standard one
    # https://www.lyonlabs.org/commodore/onrequest/geos/geos-fonts.html
    # 0x61 / 0x15d
    O_GHSETLEN=le.uint16.array(15),
    skip='uint8',
    # 0x80 / 0x17c
    O_GHFONTID='uint16',
    # 0x82 / 0x17e
    O_GHPTSIZES=le.uint16.array(15),
    # 0xa0 / 0x19c

    # > A0-FF: Description (terminated with a $00)
    description='96s',
)

# font_id values
# http://www.zimmers.net/geos/docs/fontfile.txt
# 0    BSW            13   Tilden
# 1    University     14   Evans
# 2    California     15   Durant
# 3    Roma           16   Telegraph
# 4    Dwinelle       17   Superb
# 5    Cory           18   Bowditch
# 6    Tolman         19   Ormond
# 7    Bubble         20   Elmwood
# 8    Fontknox       21   Hearst
# 9    Harmon         21   Brennens (BUG)
# 10   Mykonos        23   Channing
# 11   Boalt          24   Putnam
# 12   Stadium        25   LeConte


# >   If the file is a VLIR, then the RECORD block is of interest. This  single
# > sector is made up of up to 127 track/sector pointers, each of  which  point
# > to program sections (called RECORDS). VLIR files are comprised of  loadable
# > RECORDS (overlays, if you wish to use PC terminology). The first RECORD  is
# > what is always loaded first when you run that application. After that,  the
# > OS loads whatever RECORD it needs.
# >   When a T/S link of $00/$00 is encountered, we  are  at  the  end  of  the
# > RECORD block. If the T/S  link  is  a  $00/$FF,  then  the  record  is  not
# > available.

# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT
# >   Note that the RECORD block is modified  from  the  original  GEOS  entry.
# > Instead of containing the track and sector  references,  we  now  have  the
# > sector count and the size of the last sector in the chain
_RECORD_ENTRY = le.Struct(
    sector_count='uint8',
    last_size='uint8',
)
_RECORD_BLOCK = _RECORD_ENTRY.array(127)


_SEQ_SIGNATURE = b'SEQ formatted GEOS file V1.0'
_PRG_SIGNATURE = b'PRG formatted GEOS file V1.0'


@loaders.register(
    name='geos',
    patterns=('*.cvt',),
    magic=(
        Magic.offset(30) + _SEQ_SIGNATURE,
        Magic.offset(30) + _PRG_SIGNATURE,
    ),
)
def load_geos(instream, merge_mega:bool=True, extract_del:bool=False):
    """
    Load fonts from a GEOS ConVerT container.

    extract_del: extract the unused DEL glyph 0x7f (defauult: False)
    merge_mega: merge to mega font, if detected (default: True)
    """
    dir_entry = _DIR_BLOCK.read_from(instream)
    logging.debug('directory entry: %s',  dir_entry)
    sig_block = _SIG_BLOCK.read_from(instream)
    logging.debug('signature: %s', sig_block)
    if sig_block.signature not in (_SEQ_SIGNATURE, _PRG_SIGNATURE):
        raise FileFormatError(
            'Not a GEOS font file: incorrect signature '
            f'{sig_block.signature.decode("latin-1")}'
        )
    info_block = _INFO_BLOCK.read_from(instream)
    logging.debug('info block: %s', info_block)
    record_block = _RECORD_BLOCK.read_from(instream)
    logging.debug(
        'record block: %s',
        list(x for x in record_block if not bytes(x) == b'\0\xff')
    )
    if dir_entry.geos_filetype != _GEOS_FONT_TYPE:
        raise FileFormatError(
            'Not a GEOS font file: incorrect filetype '
            f'{dir_entry.geos_filetype:02x}'
        )
    # properties which don't change within the family
    family = _str_from_geos(dir_entry.filename.rstrip(b'\xa0'))
    class_text = _str_from_geos(info_block.class_text)
    if class_text.startswith(family):
        name, _, revision = class_text.partition('V')
        if name.strip() == family:
            class_text = None
    else:
        revision = None
    props = dict(
        family=family,
        revision=revision,
        # display icon in comment
        comment=Raster.from_bytes(tuple(info_block.icon), width=24).as_text(),
        notice=_str_from_geos(info_block.description),
    )
    props['geos.class_text'] = class_text
    props['geos.timestamp'] = (
        f'{dir_entry.year+1900:04d}-{dir_entry.month:02d}-{dir_entry.day:02d} '
        f'{dir_entry.hour:02d}:{dir_entry.minute:02d}'
    )
    # create fonts
    fonts = {}
    for data_size, ghptsize in zip(
            info_block.O_GHSETLEN,
            info_block.O_GHPTSIZES
        ):
        if not ghptsize or not data_size:
            continue
        # ptsize is (font_id << 6) + point_size
        font_id, height = divmod(ghptsize, 1<<6)
        logging.debug('Loading font id %d index %d', font_id, height)
        anchor = instream.tell()
        try:
            font = load_geos_vlir_record(
                Stream(instream, mode='r'), extract_del=extract_del
            )
        except ValueError as e:
            logging.warning(
                'Could not load font id %d size %d: %s', font_id, height, e
            )
            # flag to skip to next sector
            font = None
        true_data_size = instream.tell() - anchor
        if true_data_size != data_size:
            logging.warning(
                'Actual size 0x%x differs from reported size 0x%x',
                true_data_size, data_size
            )
        # go to end of sector
        # 254 bytes per sector - the cvt does not store the initial pointer
        nxt = ceildiv(true_data_size, 254) * 254
        instream.seek(anchor + nxt)
        if font is not None:
            font = font.modify(font_id=font_id, **props)
            fonts[height] = font
    if merge_mega and _is_mega(fonts):
        return _merge_mega(fonts)
    else:
        return tuple(fonts.values())


def _is_mega(fonts):
    """Check if extracted fonts represent one mega font."""
    # mega fonts: glyphs are divided over multiple strikes
    # undefined glyphs are given as 1-pixel-wide
    # GHPTSIZE differs between strikes but actual pixel-size is the same
    # sometimes the last (empty) strike has a different height
    id_sizes = set(
        (_f.font_id, _f.pixel_size)
        for _index, _f in fonts.items() if _index != 54
    )
    # if 7 strikes exist and 6 have the same id and pixel_size, assume mega font
    return len(id_sizes) == 1 and tuple(fonts.keys()) == tuple(range(48, 55))


def _merge_mega(fonts):
    """Merge fonts to mega font."""
    logging.info('Mega font detected, merging.')
    fonts = tuple(fonts.values())
    # last strike should contain only empty or blank glyphs
    last_empty = all(_g.is_blank() for _g in fonts[-1].glyphs)
    if not last_empty:
        logging.warning(
            'Last strike in mega font is not empty: glyphs will be discarded.'
        )
    # take glyph of maximum width from each strike
    selected = tuple(
        max(glyphs, key=lambda _g: _g.width)
        for glyphs in zip(*(_f.glyphs for _f in fonts))
    )
    if len(set(_g.height for _g in selected)) > 1:
        logging.warning('Mega font strikes differ in height')
    if len(set(_g.shift_up for _g in selected)) > 1:
        logging.warning('Mega font strikes differ in baseline')
    fonts = (fonts[0].modify(glyphs=selected),)
    return fonts


def _str_from_geos(text):
    """Convert string from GEOS format."""
    return text.decode('ascii', 'replace').replace('\r', '\n')


###############################################################################

@savers.register(linked=load_geos)
def save_geos(fonts, outstream, *, mega:bool=False):
    """
    Save fonts to a GEOS ConVerT container.

    mega: save in mega font format (single font only; default: False)
    """
    fonts = tuple(_prepare_font_for_geos_vlir_record(_f) for _f in fonts)
    if not mega:
        logging.debug('Creating GEOS regular font.')
        subfonts, common_props = _prepare_geos(fonts)
    else:
        logging.debug('Creating GEOS mega font.')
        subfonts, common_props = _split_mega(fonts)
    _write_geos(subfonts, outstream, **common_props)


def _write_geos(
        subfonts, outstream,
        family, revision, font_id,
        notice, class_text, timestamp
    ):
    """Write out prepared fonts to GEOS format."""
    records = {
        _index: _create_geos_vlir_record(_f)
        for _index, _f in subfonts.items()
    }
    if timestamp:
        dt = datetime.fromisoformat(timestamp)
        year, month, day, hour, min, *_ = dt.timetuple()
    else:
        year, month, day, hour, min = 1900, 1, 1, 0, 0
    dir_block = _DIR_BLOCK(
        # C64 USR file (0x03), closed (0x80)
        filetype=0x83,
        # no idea if this is a legal value
        sector=0x100,
        filename=_str_to_geos(family).ljust(16, b'\xa0'),
        # sems to be always one less in the hi byte than 'sector' above
        info_sector=0,
        # VLIR structure (Sequential is 0x00)
        geos_structure=0x01,
        # > GEOS filetype
        geos_filetype=_GEOS_FONT_TYPE,
        year=year-1900,
        month=month,
        day=day,
        hour=hour,
        minute=min,
        # filesize='uint16',
    )
    sig_block = _SIG_BLOCK(
        # "SEQ" also happens, but "PRG" is more common (yet filetype is USR)
        signature=_PRG_SIGNATURE,
        # notes field is "usually $00"
    )
    # create the info block
    info_block = _INFO_BLOCK(
        id_bytes=b'\x03\x15\xBF',
        # > Icon bitmap (sprite format, 63 bytes)
        icon=(le.uint8 * 63)(*_make_icon()),
        filetype=dir_block.filetype,
        geos_filetype=dir_block.geos_filetype,
        geos_structure=dir_block.geos_structure,
        # 3 addresses based on sample file
        load_address=0,
        end_address=0xffff,
        start_address=0,
        class_text=_str_to_geos(class_text) or _make_classtext(family, revision),
        O_GHFONTID=font_id,
        # ptsize is (font_id << 6) + point_size
        O_GHPTSIZES=(le.uint16 * 15)(*((font_id << 6) + _r for _r in records)),
        # the size in bytes of the corresponding VLIR record when loaded
        O_GHSETLEN=(le.uint16 * 15)(*(len(_r) for _r in records.values())),
        description=(_str_to_geos(notice) or _make_description(subfonts.keys())),
    )
    # create the record block
    empty = _RECORD_ENTRY(sector_count=0, last_size=0xff,)
    entries = [empty] * 127
    # entry index corresponds with font's pixel size
    for pixel_size, rec in records.items():
        entries[pixel_size] = _RECORD_ENTRY(
            sector_count=ceildiv(len(rec)+1, 254),
            last_size=(len(rec)+1) % 254,
        )
    record_block = _RECORD_BLOCK(*entries)
    # info block + record block + data blocks
    dir_block.filesize = 2 + sum(_e.sector_count for _e in entries)
    # write out
    outstream.write(bytes(dir_block))
    outstream.write(bytes(sig_block))
    outstream.write(bytes(info_block))
    outstream.write(bytes(record_block))
    for rec in records.values():
        outstream.write(rec)
        outstream.write(bytes((254 - len(rec)) % 254))


def _prepare_geos(fonts):
    """Validate fonts for storing in GEOS convert format; extract metadata."""
    if len(set(_f.family for _f in fonts)) > 1:
        raise UnsupportedError(
            'GEOS font file can only store fonts from one family.'
        )
    if len(set(_f.pixel_size for _f in fonts)) != len(fonts):
        raise UnsupportedError(
            'GEOS font file can only store fonts with distinct pixel sizes.'
        )
    if len(fonts) > 15:
        raise UnsupportedError(
            'GEOS font file can only store at most 15 pixel sizes.'
        )
    if max(_f.pixel_size for _f in fonts) > 63:
        raise UnsupportedError(
            'GEOS font file can only store fonts of up to 63 pixels tall.'
        )
    common_props = _get_metadata(fonts[0])
    # sort in ascending order of pixel size
    fonts = tuple(sorted(fonts, key=lambda _f: _f.pixel_size))
    # index fonts by size
    subfonts = {_f.pixel_size: _f for _f in fonts}
    return subfonts, common_props


# https://www.lyonlabs.org/commodore/onrequest/geos/geos-fonts.html
# > 48: $20-$2f (blank to '/')
# > 49: $30-$3f ('0' to '?')
# > 50: $40-$4f ('@' to 'O')
# > 51: $50-$5f ('P' to '_')
# > 52: $60-$6f ('`' to 'o')
# > 53: $70-$7f ('p' to DEL)
_MEGA_SUBSETS = {
    48: range(0x20, 0x30),
    49: range(0x30, 0x40),
    50: range(0x40, 0x50),
    51: range(0x50, 0x60),
    52: range(0x60, 0x70),
    53: range(0x70, 0x80),
}

_GEOS_RANGE = range(0x20, 0x80)


def _split_mega(fonts):
    """Prepare font for storing in mega format."""
    font = ensure_single(fonts)
    common_props = _get_metadata(font)
    subfonts = {
        _index: font.subset(codepoints=_subrange)
        for _index, _subrange in _MEGA_SUBSETS.items()
    }
    # > the point sizes containing partial character sets contain proper
    # > bitstream indices for each character they contain, but an offset of
    # > one pixel for the characters that do not appear in that record.
    placeholder = Glyph.blank(width=1, height=font.glyphs[0].height, shift_up=font.glyphs[0].shift_up)
    subfonts = {
        _index: _subfont.resample(codepoints=_GEOS_RANGE, missing=placeholder)
        for _index, _subfont in subfonts.items()
    }
    # > There is also a 54-point record, which contains a complete bitstream
    # > index for every character (e.g. the indices for '0' to '?' are the same
    # > as those in record 49), but no bitstream data
    subfonts[54] = Font(Glyph(codepoint=_c) for _c in _GEOS_RANGE)
    return subfonts, common_props


def _get_metadata(font):
    """Get font family metadata."""
    family = font.family[:15]
    revision = font.revision
    try:
        font_id = int(font.font_id)
    except ValueError:
        font_id = 1023
    if 0 > font_id > 1023:
        font_id = 1023
    notice = font.notice[:95]
    return dict(
        family=family,
        revision=revision,
        font_id=font_id,
        notice=notice or '',
        class_text=font.get_property('geos.class_text') or '',
        timestamp=font.get_property('geos.timestamp'),
    )


def _str_to_geos(text):
    """Convert string to GEOS format."""
    return text.replace('\n', '\r').encode('ascii', 'replace')


def _make_description(sizes):
    """Create standard description string."""
    sizes = tuple(sizes)
    # actually pixels not points
    pointlist = ', '.join(str(_size) for _size in sizes[:-1])
    if pointlist:
        pointlist += ' and '
    pointlist += str(sizes[-1])
    description = f'Available in {pointlist} point.'
    return _str_to_geos(description)[:95]


def _make_classtext(family, revision):
    """Create standard class text."""
    # 20 bytes, with null terminator
    if revision != '0':
        revision = revision[:3]
        text = f'{family.ljust(11)} V{revision.ljust(3)}'
    else:
        text = f'{family}'
    return _str_to_geos(text)


_FONT_ICON = """\
@@@@@@@@@@@@@@@@@@@@@@@@
@......................@
@......................@
@......................@
@.@@@@@@@..............@
@..@@...@...........@..@
@..@@..............@@..@
@..@@..............@@..@
@..@@@@...........@@@@.@
@..@@...@@@..@@@@..@@..@
@..@@..@@.@@.@@.@@.@@..@
@..@@..@@.@@.@@.@@.@@..@
@..@@..@@.@@.@@.@@.@@..@
@.@@@@..@@@..@@.@@..@@.@
@......................@
@......................@
@......................@
@......................@
@......................@
@......................@
@@@@@@@@@@@@@@@@@@@@@@@@
"""

def _make_icon():
    """Create standard icon."""
    icon = Raster.from_vector(_FONT_ICON, stride=25, width=24, inklevels='.@')
    return icon.as_bytes()



"""
monobit.storage.fontformats.raw.grasp - GRASP / PCPaint character-cell "old format"

(c) 2022--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers
from monobit.base import FileFormatError
from monobit.core import Font, Glyph, Raster
from monobit.base.struct import little_endian as le
from monobit.base.binary import ceildiv
from .raw import load_bitmap, save_bitmap

from monobit.storage.utils.limitations import ensure_single, ensure_charcell, make_contiguous


###############################################################################
# PCPaint / GRASP original format. Not used by FONTRIX or ChiWriter.
# suffix .SET or .FNT
#
# http://fileformats.archiveteam.org/wiki/GRASP_font
# http://www.textfiles.com/programming/FORMATS/glformat.txt
#
# +-- Font Header
# | length    (word)    length of the entire font file
# | size      (byte)    number of glyphs in the font file
# | first     (byte)    byte value represented by the first glyph
# | width     (byte)    width of each glyph in pixels
# | height    (byte)    height of each glyph in pixels
# | glyphsize (byte)    number of bytes to encode each glyph
# +-- Glyph Data

_GRASP_HEADER = le.Struct(
    filesize='uint16',
    count='uint8',
    first='uint8',
    width='uint8',
    height='uint8',
    glyphsize='uint8',
)


@loaders.register(
    name='grasp',
    patterns=(
        '*.set', '*.fnt',
    ),
)
def load_grasp_old(instream):
    """
    Load a GRASP/PCPaint character-cell "old format" font.
    This is not the FONTRIX-based proportional "new format".
    """
    header = _GRASP_HEADER.read_from(instream)
    logging.debug('GRASP old-format header: %s', header)
    if header.filesize & 0xF in (0x10, 0x11):
        # 7-byte header, usually binary-rounded character cells & counts
        # so first byte for old format is usualy 0x07 or 0x87
        logging.warning(
            'First byte consistent with new-format GRASP font.'
        )
    if (0 in (header.count, header.glyphsize, header.width, header.height)):
        raise FileFormatError('Bad geometry for old-format GRASP font.')
    font = load_bitmap(
        instream,
        width=header.width, height=header.height,
        # strike_bytes=header.glyphsize // header.height,
        strike_count=1,
        count=header.count,
        first_codepoint=header.first,
    )
    font = font.modify(source_format='Old-format GRASP font')
    return font


@savers.register(linked=load_grasp_old)
def save_grasp_old(fonts, outstream):
    """
    Save to GRASP/PCPaint character-cell "old format" font.
    This is not the FONTRIX-based proportional "new format".
    """
    font = ensure_single(fonts)
    font = ensure_charcell(font)
    # header.count is a 1-byte field which limits us to 255 glyphs
    font = make_contiguous(font, supported_range=range(255), missing='space')
    header = _GRASP_HEADER(
        count=len(font.glyphs),
        first=int(min(font.get_codepoints())),
        width=font.cell_size.x,
        height=font.cell_size.y,
        glyphsize=ceildiv(font.cell_size.x, 8) * font.cell_size.y,
    )
    header.filesize = header.count * header.glyphsize + _GRASP_HEADER.size
    outstream.write(bytes(header))
    save_bitmap(outstream, font)



f"""
monobit.storage.fontformats.fontx - DOS/V FONTX2 format

(c) 2022--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate

from monobit.base.struct import bitfield, little_endian as le
from monobit.base.binary import ceildiv
from monobit.base import Props, FileFormatError
from monobit.storage import loaders, savers
from monobit.core import Font, Glyph

from monobit.storage.utils.limitations import ensure_single, ensure_charcell


_FONTX_MAGIC = b'FONTX2'


@loaders.register(
    name='fontx',
    magic=(_FONTX_MAGIC,),
    patterns=('*.fnt',),
)
def load_fontx(instream):
    """Load font from fontx file."""
    fontx_props, glyphs = _read_fontx(instream)
    logging.info('fontx properties:')
    for line in str(fontx_props).splitlines():
        logging.info('    ' + line)
    props = _convert_from_fontx(fontx_props)
    return Font(glyphs, **props)


@savers.register(linked=load_fontx)
def save_fontx(fonts, outstream, endianness:str='little'):
    """Save font to fontx file."""
    font = ensure_single(fonts)
    endian = endianness[0].lower()
    props, blocks, glyphs = _convert_to_fontx(font)
    logging.info('fontx properties:')
    for line in str(props).splitlines():
        logging.info('    ' + line)
    _write_fontx(outstream, props, blocks, glyphs)


################################################################################
# fontx binary format

# https://unifoundry.com/japanese/index.html
# http://elm-chan.org/docs/dosv/fontx_e.html


_SBCS_HEADER = le.Struct(
    magic='6s',
    name='8s',
    # Glyph Width (pixels), ncols
    width='uint8',
    # Glyph Height (pixels), nrows
    height='uint8',
    # Code Flag:
    #   0 for Alphanumeric Keyboard
    #   1 for Shift-JIS
    code_flag='uint8',
)
_DBCS_HEADER= le.Struct(
    # Number of Code Blocks, nb
    n_blocks='uint8',
)
_BLOCK_OFFSET = le.Struct(
    # Code Block Start
    # lowest low-byte in block
    start='uint16',
    # Code Block nb End
    # highest low-byte in block
    end='uint16',
)


def _read_fontx(instream):
    """Read fontx binary file and return as properties."""
    endian = 'l'
    sbcs_header = _SBCS_HEADER.read_from(instream)
    if sbcs_header.magic != _FONTX_MAGIC:
        raise FileFormatError(
            'Not a FONTX file: '
            f'incorrect magic `{sbcs_header.magic}` != {_FONTX_MAGIC}'
        )
    bytewidth = ceildiv(sbcs_header.width, 8)
    bytesize = bytewidth * sbcs_header.height
    props = vars(sbcs_header)
    if not sbcs_header.code_flag:
        glyphs = tuple(
            Glyph.from_bytes(
                instream.read(bytesize),
                width=sbcs_header.width,
                codepoint=_cp
            )
            for _cp in range(0, 256)
        )
    else:
        dbcs_header = _DBCS_HEADER.read_from(instream)
        props.update(vars(dbcs_header))
        blocks = _BLOCK_OFFSET.array(dbcs_header.n_blocks).read_from(instream)
        block_offs = (0, ) + tuple(accumulate(
            (_entry.end-_entry.start+1)*bytesize
            for _entry in blocks
        ))
        data = instream.read()
        glyphs = tuple(
            Glyph.from_bytes(
                data[_ofs+_i*bytesize:_ofs+(_i+1)*bytesize],
                width=sbcs_header.width,
                codepoint=_cp
            )
            for _b, (_entry, _ofs) in enumerate(zip(blocks, block_offs))
            for _i, _cp in enumerate(range(_entry.start, _entry.end+1))
        )
    return Props(**props), glyphs


def _convert_from_fontx(fontx_props):
    """Convert fontx properties and glyphs to standard."""
    props = dict(
        name=fontx_props.name.decode('latin-1', 'ignore'),
    )
    return props


###############################################################################
# writer

def _convert_to_fontx(font):
    """Convert monobit font to fontx properties and glyphs."""
    font = ensure_charcell(font)
    blank = Glyph.blank(*font.raster_size)
    props = Props(
        code_flag=len(max(font.get_codepoints())) > 1,
        name=font.name[:8].encode('ascii', 'replace'),
        width=font.cell_size.x,
        height=font.cell_size.y,
    )
    if not props.code_flag:
        # sbcs_font
        blocks = (range(256),)
    else:
        # dbcs font
        # find contiguous ranges and build block table
        # at most 256 blocks; keep one page in one block and contiguous range within
        # at most 2-byte codepoints
        cps = tuple(int(_c) for _c in font.get_codepoints())
        pages = tuple(
            tuple(
                _c for _c in cps
                if _c >= 0x100*_hi and _c < 0x100*(_hi+1)
            )
            for _hi in range(256)
        )
        # skip empty ranges
        blocks = tuple(
            range(min(_page), max(_page)+1)
            for _page in pages if _page
        )
    codepoints = (_cp for _range in blocks for _cp in _range)
    font = font.resample(codepoints=codepoints, missing=blank)
    return props, blocks, font.glyphs

def _write_fontx(outstream, props, blocks, glyphs):
    """Write fontx properties and glyphs to binary file."""
    glyph_bytes = tuple(_g.as_bytes() for _g in glyphs)
    bitmap = b''.join(glyph_bytes)
    offsets = (0,) + tuple(accumulate(len(_g) for _g in glyph_bytes))
    sbcs_header = _SBCS_HEADER(
        magic=_FONTX_MAGIC,
        **vars(props)
    )
    if props.code_flag:
        dbcs_header = _DBCS_HEADER(
            n_blocks=len(blocks)
        )
        block_offsets = _BLOCK_OFFSET.array(len(blocks))(*(
            _BLOCK_OFFSET(start=min(_block), end=max(_block))
            for _block in blocks
        ))
        outstream.write(
            bytes(sbcs_header)
            + bytes(dbcs_header)
            + bytes(block_offsets)
            + bitmap
        )
    else:
        outstream.write(
            bytes(sbcs_header)
            + bitmap
        )



"""
monobit.storage.wrapperformats.unixcompress - compress (.Z) encoding

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from io import BytesIO
from pathlib import Path

from monobit.base import safe_import
ncompress = safe_import('ncompress')

from ..streams import Stream, DelayedWriterStream
from ..base import encoders, decoders


if ncompress:

    @decoders.register(
        name='compress',
        patterns=('*.Z',),
        magic=(b'\x1f\x9d',),
    )
    def decode_compress(instream):
        data = ncompress.decompress(instream)
        name = Path(instream.name).stem
        return Stream.from_data(data, mode='r', name=name)


    @encoders.register(linked=decode_compress)
    def encode_compress(outstream):
        encode_func = ncompress.compress
        name = Path(outstream.name).stem
        return DelayedWriterStream(outstream, encode_func, name)



"""
monobit.render.glyphmap - glyph maps

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import safe_import
Image = safe_import('PIL.Image')

from monobit.base.blocks import matrix_to_blocks, matrix_to_shades, blockstr
from ..base import Props, Coord, RGB
from ..core.raster import turn_method
from ..plumbing import convert_arguments


def glyph_to_image(glyph, paper, ink):
    """Create image of single glyph."""
    image_mode = _get_image_mode(paper, ink, paper)
    charimg = Image.new(image_mode, (glyph.width, glyph.height))
    # create ink gradient
    if isinstance(ink, int):
        ink = (ink,)
        paper = (paper,)
    maxlevel = glyph.levels - 1
    inklevels = tuple(
        tuple(
            (_i * _ink + (maxlevel - _i) * _paper) // maxlevel
            for _ink, _paper in zip(ink, paper)
        )
        for _i in range(glyph.levels)
    )
    data = glyph.as_pixels(inklevels=inklevels)
    if image_mode in ('RGB', 'RGBA'):
        # itertools grouper idiom, split in groups of 3 or 4 bytes
        iterators = [iter(data)] * len(image_mode)
        # strict=True requires python 3.10 or above
        data = tuple(zip(*iterators)) #, strict=True))
    charimg.putdata(data)
    return charimg


def _get_image_mode(*colourspec):
    if not Image:
        raise ImportError('Rendering to image requires PIL module.')
    if len(set(type(_c) for _c in colourspec)) > 1:
        raise TypeError(
            'paper, ink and border must be of the same type; '
            f'got {colourspec}'
        )
    paper, ink, border = colourspec
    if paper == border == 0 and ink == 1:
        image_mode = '1'
    elif isinstance(paper, int):
        image_mode = 'L'
    elif isinstance(paper, tuple) and len(paper) == 3:
        image_mode = 'RGB'
    elif isinstance(paper, tuple) and len(paper) == 4:
        image_mode = 'RGBA'
    else:
        raise TypeError('paper, ink and border must be either int or RGB(A) tuple')
    return image_mode


class GlyphMap:

    def __init__(self, map=()):
        self._map = list(map)
        self._labels = []
        self._turns = 0
        self._scale_x = 1
        self._scale_y = 1

    def __iter__(self):
        return iter(self._map)

    def __contains__(self, item):
        return item in self._map

    def turn(self, clockwise=1, *, anti=0):
        turns = clockwise - anti
        if turns%2:
            self._scale_x, self._scale_y = self._scale_y, self._scale_x
        self._turns += turns

    def stretch(self, scale_x=1, scale_y=1):
        self._scale_x *= scale_x
        self._scale_y *= scale_y

    def append_glyph(self, glyph, x, y, sheet=0):
        """Insert glyph in glyph map."""
        self._map.append(Props(glyph=glyph, x=x, y=y, sheet=sheet))

    def append_label(self, text, x, y, *, sheet=0, right_align=False):
        """Insert glyph in glyph map."""
        self._labels.append(Props(text=text, x=x, y=y, sheet=sheet, right_align=right_align))

    def reorder(self, mapping):
        """Rearrange glyphs through index mapping."""
        self._map = [
            self._map[mapping[_i]]
            for _i in range(len(self._map))
        ]

    def get_bounds(self):
        """Get extreme coordinates across all sheets. Bounds are inclusive."""
        last = max(_entry.sheet for _entry in self._map)
        min_x = min(_entry.x for _entry in self._map)
        min_y = min(_entry.y for _entry in self._map)
        # inclusive bounds - we included the width/height of the glyphs
        # e.g. if I have a 2-pixel wide glyph at x=0, I need a 2-pixel image
        max_x = max(_entry.x + _entry.glyph.width for _entry in self._map)
        max_y = max(_entry.y + _entry.glyph.height for _entry in self._map)
        return last, min_x, min_y, max_x, max_y

    def to_canvas(self, sheet=0):
        """Convert one sheet of the glyph map to canvas."""
        # note that this gives the bounds across *all* sheets
        _, min_x, min_y, max_x, max_y = self.get_bounds()
        # no +1 as bounds are inclusive
        canvas = _Canvas.blank(max_x - min_x, max_y - min_y)
        for entry in self._map:
            if entry.sheet == sheet:
                canvas.blit(entry.glyph, entry.x - min_x, entry.y - min_y)
        for entry in self._labels:
            if entry.sheet == sheet:
                canvas.write(entry.text, entry.x - min_x, entry.y - min_y, entry.right_align)
        return canvas.stretch(self._scale_x, self._scale_y).turn(self._turns)

    def to_images(
            self, *, paper=0, ink=255, border=0, invert_y=False,
            transparent=True,
        ):
        """Draw images based on sheets in glyph map."""
        image_mode = _get_image_mode(paper, ink, border)
        last, min_x, min_y, max_x, max_y = self.get_bounds()
        # no +1 as bounds are inclusive
        width, height = max_x - min_x, max_y - min_y
        images = [Image.new(image_mode, (width, height), border) for _ in range(last+1)]
        for entry in self._map:
            # we need to treat the background colour as transparent
            # we create the glyph as a mask and cut it from a solid ink-colour
            if transparent:
                mask = glyph_to_image(entry.glyph, 0, 255)
                colour = Image.new(image_mode, mask.size, ink)
            else:
                mask = None
                colour = glyph_to_image(entry.glyph, paper, ink)
            if invert_y:
                target = (entry.x, entry.y)
            else:
                # Image has ttb y coords, we have btt
                # our character origin is bottom left
                target = (entry.x-min_x, height-entry.glyph.height+min_y-entry.y)
            images[entry.sheet].paste(colour, target, mask)
        images = tuple(
            _im.resize((self._scale_x*_im.width, self._scale_y*_im.height))
                .rotate(-90 * self._turns, expand=True)
            for _im in images
        )
        return images

    def as_image(
            self, *,
            ink=255, paper=0, border=0,
            sheet=0, invert_y=False,
        ):
        """Convert glyph map to image."""
        images = self.to_images(
            ink=ink, paper=paper, border=border, invert_y=invert_y,
        )
        return images[sheet]

    def as_text(
            self, *,
            inklevels=' @',
            border=None,
            start='', end='\n',
            sheet=0,
        ):
        """Convert glyph map to text."""
        canvas = self.to_canvas(sheet=sheet)
        return canvas.as_text(
            inklevels=inklevels, border=border, start=start, end=end
        )

    @convert_arguments
    def as_blocks(self, resolution:Coord=Coord(2, 2), *, sheet=0):
        """Convert glyph map to a string of quadrant block characters."""
        canvas = self.to_canvas(sheet=sheet)
        return canvas.as_blocks(resolution)

    def as_shades(
            self, *,
            paper=RGB(0, 0, 0), ink=RGB(255, 255, 255), border=None, sheet=0,
        ):
        """Convert glyph map to ansi coloured block characters."""
        canvas = self.to_canvas(sheet=sheet)
        return canvas.as_shades(paper=paper, ink=ink, border=border)

    def get_sheet(self, sheet=0):
        """Return glyph records for a given sheet."""
        return tuple(_e for _e in self._map if _e.sheet == sheet)

    def get_sheet_labels(self, sheet=0):
        """Return labels for a given sheet."""
        return tuple(_e for _e in self._labels if _e.sheet == sheet)


class _Canvas:
    """Blittable raster for glyph maps."""

    def __init__(self, pixels, levels=2, labels=()):
        """Create raster from tuple of tuples of string."""
        self._pixels = pixels
        self._labels = list(labels)
        self.height = len(pixels)
        self.width = 0 if not pixels else len(pixels[0])
        self.levels = levels

    @classmethod
    def blank(cls, width, height, fill=-1):
        """Create a canvas in background colour."""
        canvas = [[fill]*width for _ in range(height)]
        return cls(canvas)

    def blit(self, raster, grid_x, grid_y):
        """Draw a matrix onto a canvas, leaving existing ink in place."""
        if not raster.width or not self.width:
            return self
        matrix = raster.as_matrix()
        self.levels = max(self.levels, raster.levels)
        for work_y in reversed(range(raster.height)):
            if 0 <= grid_y + work_y < self.height:
                row = self._pixels[self.height - (grid_y + work_y) - 1]
                for work_x, ink in enumerate(matrix[raster.height - work_y - 1]):
                    if 0 <= grid_x + work_x < self.width:
                        # grayscale will be additive until full-ink level
                        row[grid_x + work_x] = min(
                            self.levels - 1,
                            max(0, ink) + max(0, row[grid_x + work_x]),
                        )
        return self

    def write(self, text, x, y, right_align=False):
        """Add a text label onto the canvas"""
        self._labels.append((text, x, y, right_align))

    def _write_labels_to_matrix(self, matrix, resolution=Coord(1, 1)):
        """Write labels to text or blocks matrix."""
        if not matrix:
            return
        for text, x, y, right_align in self._labels:
            x //= resolution.x
            y //= resolution.y
            if y < 0 or y > len(matrix) - 1:
                continue
            width = len(matrix[0])
            text = list(text)
            if right_align:
                if x - len(text) < 0:
                    text = text[:-x+len(text)]
                matrix[len(matrix) - y - 1][x-len(text) : x] = text
            else:
                if x + len(text) > self.width:
                    text = text[:self.width-x-len(text)]
                matrix[len(matrix) - y - 1][x : x+len(text)] = text

    def as_text(
            self, *,
            inklevels=' @', border=None,
            start='', end='\n'
        ):
        """Convert glyph map to text."""
        if not self.height:
            return ''
        if self.levels > len(inklevels):
            raise ValueError(f'Requires at least {self.levels} greyscale levels.')
        if not border:
            border = inklevels[0]
        colourdict = {-1: border} | {
            _i: _v for _i, _v in enumerate(inklevels)
        }
        matrix = [
            [colourdict[_pix] for _pix in _row]
            for _row in self._pixels
        ]
        # write out labels
        self._write_labels_to_matrix(matrix)
        # join all text together
        contents = '\n'.join(''.join(_row) for _row in matrix)
        return blockstr(''.join((start, contents, end)))

    def as_blocks(self, resolution=Coord(2, 2)):
        """Convert glyph map to a string of block characters."""
        if not self.height:
            return ''
        # replace background with paper
        pixels = tuple(
            tuple(
                0 if _pix == -1 else _pix
                for _pix in _row
            )
            for _row in self._pixels
        )
        block_matrix = matrix_to_blocks(pixels, *resolution, levels=self.levels)
        self._write_labels_to_matrix(block_matrix, resolution=resolution)
        blocks = '\n'.join(''.join(_row) for _row in block_matrix)
        return blockstr(blocks + '\n')

    def as_shades(self, *, paper, ink, border):
        """Convert glyph map to a string of block characters with ansi colours."""
        if not self.height:
            return ''
        block_matrix = matrix_to_shades(
            self._pixels, levels=self.levels,
            paper=paper, ink=ink, border=border,
        )
        self._write_labels_to_matrix(block_matrix)
        blocks = '\n'.join(''.join(_row) for _row in block_matrix)
        return blockstr(blocks + '\n')

    def stretch(self, factor_x:int=1, factor_y:int=1):
        """
        Repeat rows and/or columns.

        factor_x: number of times to repeat horizontally
        factor_y: number of times to repeat vertically
        """
        # vertical stretch
        pixels = (_row for _row in self._pixels for _ in range(factor_y))
        # horizontal stretch
        pixels = [
            [_col for _col in _row for _ in range(factor_x)]
            for _row in pixels
        ]
        # adjust labels
        labels = [
            (_text, _x*factor_x, _y*factor_y, _ralign)
            for _text, _x, _y, _ralign in self._labels
        ]
        return type(self)(pixels, levels=self.levels, labels=labels)

    def flip(self):
        """Reverse pixels vertically."""
        # adjust labels
        labels = [
            (_text, _x, self.height-_y-1, _ralign)
            for _text, _x, _y, _ralign in self._labels
        ]
        return type(self)(self._pixels[::-1], levels=self.levels, labels=labels)

    def transpose(self):
        """Transpose glyph."""
        # adjust labels
        labels = [
            (_text, _y, _x, _ralign)
            for _text, _x, _y, _ralign in self._labels
        ]
        return type(self)(
            [list(_r) for _r in zip(*self._pixels)],
            levels=self.levels,
            labels=labels,
        )

    turn = turn_method



"""
monobit.storage.containerformats.ace - ace aechive reader

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import PurePosixPath

from monobit.base import safe_import
acefile = safe_import('acefile')

from monobit.base import FileFormatError
from ..magic import Magic
from ..base import containers
from ..streams import Stream
from ..containers import Archive


if acefile:

    @containers.register(
        name='ace',
        patterns=('*.ace',),
        magic=(
            Magic.offset(7) + b'**ACE**',
        ),
    )
    class AceContainer(Archive):
        """Container for ACE archives."""

        def __init__(self, stream, mode='r'):
            """Read ACE archives."""
            if mode != 'r':
                raise ValueError('Writing to ACE archives not supported')
            super().__init__(stream, mode)
            self._archive = acefile.open(stream)
            self._entries = tuple(self._archive)

        def close(self):
            """Close the archive."""
            if self.closed:
                return
            try:
                self._archive.close()
            except EnvironmentError as e:
                logging.debug(e)
            self._entries = ()
            super().close()

        def list(self):
            entries =  tuple(
                f'{_e.filename}/' if _e.is_dir() else _e.filename
                for _e in self._entries
            )
            return entries

        def decode(self, name, *, password:bytes=None):
            """
            Extract file from ACE archive.

            password: password for encrypted archive entry.
            """
            filename = str(PurePosixPath(self.root) / name)
            try:
                for entry in self._entries:
                    if entry.filename == filename:
                        if entry.is_dir():
                            raise IsADirectoryError(
                                f"Entry '{filename}' is a directory in archive {self}"
                            )
                        data = self._archive.read(entry, pwd=password)
                        return Stream.from_data(data, mode='r', name=name)
            except acefile.AceError as e:
                raise FileFormatError(e) from e
            raise FileNotFoundError(
                f"Entry '{filename}' not found in archive {self}"
            )

        def encode(self, name):
            """Writing to ACE archives not supported."""
            raise ValueError(f'Writing to ACE archives not supported.')



"""
monobit.storage.fontformats.fon.windows.fnt - Windows FNT resource

`monobit.storage.fontformats.fon.windows` is copyright 2019--2024 Rob Hagemans
`mkwinfont` is copyright 2001 Simon Tatham. All rights reserved.
`dewinfont` is copyright 2001,2017 Simon Tatham. All rights reserved.

See `LICENSE.md` in this package's directory.
"""


import io
import string
import logging
import itertools
from types import SimpleNamespace

from monobit.base.binary import ceildiv, align
from monobit.base.struct import little_endian as le
from monobit.storage import loaders, savers
from monobit.base import reverse_dict, FileFormatError, UnsupportedError, Props
from monobit.core import Font, Glyph, Raster, StrokePath
from monobit.storage.utils.limitations import ensure_single

# tables used by other formats
from ...common import (
    WEIGHT_MAP, WEIGHT_REVERSE_MAP,
    CHARSET_MAP, CHARSET_REVERSE_MAP
)
from monobit.storage.utils.limitations import make_contiguous


FNT_MAGIC_1 = b'\0\1'
FNT_MAGIC_2 = b'\0\2'
FNT_MAGIC_3 = b'\0\3'


@loaders.register(
    name='win',
    magic=(FNT_MAGIC_1, FNT_MAGIC_2, FNT_MAGIC_3),
    patterns=('*.fnt',),
)
def load_win_fnt(instream):
    """Load font from a Windows .FNT resource."""
    resource = instream.read()
    font = convert_win_fnt_resource(resource)
    return font


@savers.register(linked=load_win_fnt)
def save_win_fnt(fonts, outstream, version:int=2, vector:bool=False):
    """
    Save font to a Windows .FNT resource.

    version: Windows font format version (default 2)
    vector: output a vector font (if the input font has stroke paths defined; default False)
    """
    font = ensure_single(fonts)
    outstream.write(create_fnt(font, version*0x100, vector))
    return font



##############################################################################
# windows .FNT format definitions
#
# https://web.archive.org/web/20120215123301/http://support.microsoft.com/kb/65123
# https://ffenc.blogspot.com/2008/04/fnt-font-file-format.html


# fallback values for font file writer
# use OEM charset value; "default" charset 0x01 is not a valid value per freetype docs
_FALLBACK_CHARSET = 0xff
# "dfDefaultChar should indicate a special character in the font which is not a space."
# codepoint 0x80 is unmapped in windows-ansi-2.0 and commonly used for default
_FALLBACK_DEFAULT = 0x80
# "dfBreakChar is normally (32 - dfFirstChar), which is an ASCII space."
_FALLBACK_BREAK = 0x20

# despite the presence of MBCS character set codes,
# FNT can only hold single-byte codepoints
# as firstChar and lastChar are byte-sized
_FNT_RANGE = range(256)


# pitch and family
# low bit: 1 - proportional 0 - monospace
# upper bits: family (like bdf add_style_name)
#
# Don't care or don't know.
_FF_DONTCARE = 0<<4
# Proportionally spaced fonts with serifs.
_FF_ROMAN = 1<<4
# Proportionally spaced fonts without serifs.
_FF_SWISS = 2<<4
# Fixed-pitch fonts.
_FF_MODERN = 3<<4
_FF_SCRIPT = 4<<4
_FF_DECORATIVE = 5<<4
# map to yaff styles
_STYLE_MAP = {
    _FF_DONTCARE: '',
    _FF_ROMAN: 'serif',
    _FF_SWISS: 'sans serif',
    _FF_MODERN: 'modern',
    _FF_SCRIPT: 'script',
    _FF_DECORATIVE: 'decorative',
}
_STYLE_REVERSE_MAP = reverse_dict(_STYLE_MAP)

# dfFlags
_DFF_FIXED = 0x01 # font is fixed pitch
_DFF_PROPORTIONAL = 0x02 # font is proportional pitch
_DFF_ABCFIXED = 0x04 # font is an ABC fixed font
_DFF_ABCPROPORTIONAL = 0x08 # font is an ABC proportional font
_DFF_1COLOR = 0x10 # font is one color
_DFF_16COLOR = 0x20 # font is 16 color
_DFF_256COLOR = 0x40 # font is 256 color
_DFF_RGBCOLOR = 0x80 # font is RGB color
# convenience
_DFF_PROP = _DFF_PROPORTIONAL | _DFF_ABCPROPORTIONAL
_DFF_COLORFONT = _DFF_16COLOR | _DFF_256COLOR | _DFF_RGBCOLOR
_DFF_ABC = _DFF_ABCFIXED | _DFF_ABCPROPORTIONAL


# dfType field values
# vector font, else raster
_FNT_TYPE_VECTOR = 0x0001
# font is in ROM
_FNT_TYPE_MEMORY = 0x0004
# 'realised by a device' - maybe printer font?
_FNT_TYPE_DEVICE = 0x0080


# FNT header - the part common to v1.0, v2.0, v3.0
_FNT_HEADER = le.Struct(
    ### this part is also common to FontDirEntry
    dfVersion='word',
    dfSize='dword',
    dfCopyright='60s',
    dfType='word',
    dfPoints='word',
    dfVertRes='word',
    dfHorizRes='word',
    dfAscent='word',
    dfInternalLeading='word',
    dfExternalLeading='word',
    dfItalic='byte',
    dfUnderline='byte',
    dfStrikeOut='byte',
    dfWeight='word',
    dfCharSet='byte',
    dfPixWidth='word',
    dfPixHeight='word',
    dfPitchAndFamily='byte',
    dfAvgWidth='word',
    dfMaxWidth='word',
    dfFirstChar='byte',
    dfLastChar='byte',
    dfDefaultChar='byte',
    dfBreakChar='byte',
    dfWidthBytes='word',
    dfDevice='dword',
    dfFace='dword',
###
    dfBitsPointer='dword',
    dfBitsOffset='dword',
)

# version-specific header extensions
_FNT_HEADER_1 = le.Struct()
_FNT_HEADER_2 = le.Struct(dfReserved='byte')
_FNT_HEADER_3 = le.Struct(
    dfReserved='byte',
    dfFlags='dword',
    dfAspace='word',
    dfBspace='word',
    dfCspace='word',
    dfColorPointer='dword',
    dfReserved1='16s',
)
_FNT_HEADER_EXT = {
    0x100: _FNT_HEADER_1,
    0x200: _FNT_HEADER_2,
    0x300: _FNT_HEADER_3,
}
# total size
# {'0x100': '0x75', '0x200': '0x76', '0x300': '0x94'}
_FNT_HEADER_SIZE = {
    _ver: _FNT_HEADER.size + _header.size
    for _ver, _header in _FNT_HEADER_EXT.items()
}


# GlyphEntry structures for char table
# see e.g. https://web.archive.org/web/20120215123301/http://support.microsoft.com/kb/65123
_GLYPH_ENTRY_1 = le.Struct(
    geOffset='word',
)
_GLYPH_ENTRY_2 = le.Struct(
    geWidth='word',
    geOffset='word',
)
_GLYPH_ENTRY_3 = le.Struct(
    geWidth='word',
    geOffset='dword',
)
# for ABCFIXED and ABCPROPORTIONAL; for reference, not used in v3.00 (i.e. not used at all)
_GLYPH_ENTRY_3ABC = le.Struct(
    geWidth='word',
    geOffset='dword',
    geAspace='dword',
    geBspace='dword',
    geCspace='word',
)
_GLYPH_ENTRY = {
    0x100: _GLYPH_ENTRY_1,
    0x200: _GLYPH_ENTRY_2,
    0x300: _GLYPH_ENTRY_3,
}

# proportional vector font
# these have width and offset swapped compared to the v2 bitmap format
_GLYPH_ENTRY_PVECTOR = le.Struct(
    geOffset='word',
    geWidth='word',
)


##############################################################################
# windows .FNT reader

def convert_win_fnt_resource(data):
    """Convert Windows FNT resource to monobit Font."""
    win_props = _extract_win_props(data)
    properties = _convert_win_props(data, win_props)
    glyphs = _extract_glyphs(data, win_props)
    font = Font(glyphs, **properties)
    font = font.label()
    return font

def _extract_win_props(data):
    """Read the header information in the FNT resource."""
    header = _FNT_HEADER.from_bytes(data)
    try:
        header_ext_type = _FNT_HEADER_EXT[header.dfVersion]
    except KeyError:
        raise FileFormatError(
            'Not a Windows .FNT resource or unsupported version'
            f' (0x{header.dfVersion:04x}).'
        ) from None
    extension = header_ext_type.from_bytes(data, _FNT_HEADER.size)
    win_props = SimpleNamespace(**vars(header), **vars(extension))
    return win_props

def _extract_glyphs(data, win_props):
    """Read a WinFont character table."""
    if win_props.dfType & 1:
        return _extract_vector_glyphs(data, win_props)
    if win_props.dfVersion == 0x100:
        return _extract_glyphs_v1(data, win_props)
    return _extract_glyphs_v2(data, win_props)

def _extract_glyphs_v1(data, win_props):
    """Read a WinFont 1.0 character table."""
    n_chars = win_props.dfLastChar - win_props.dfFirstChar + 1
    if not win_props.dfPixWidth:
        # proportional font
        ct_start = _FNT_HEADER_SIZE[win_props.dfVersion]
        glyph_entry_array = _GLYPH_ENTRY[win_props.dfVersion].array(n_chars+1)
        entries = glyph_entry_array.from_bytes(data, ct_start)
        offsets = [_entry.geOffset for _entry in entries]
    else:
        offsets = [
            win_props.dfPixWidth * _ord
            for _ord in range(n_chars+1)
        ]
    bytewidth = win_props.dfWidthBytes
    offset = win_props.dfBitsOffset
    strike = Raster.from_bytes(
        data[offset : offset+win_props.dfPixHeight*bytewidth],
        height=win_props.dfPixHeight,
    )
    glyphs = []
    for ord in range(n_chars):
        offset = offsets[ord]
        width = offsets[ord+1] - offset
        if not width:
            continue
        raster = strike.crop(left=offset, right=strike.width-offset-width)
        glyph = Glyph(raster, codepoint=win_props.dfFirstChar + ord)
        glyphs.append(glyph)
    return glyphs

def _extract_glyphs_v2(data, win_props):
    """Read a WinFont 2.0 or 3.0 character table."""
    n_chars = win_props.dfLastChar - win_props.dfFirstChar + 1
    glyph_entry_array = _GLYPH_ENTRY[win_props.dfVersion].array(n_chars)
    ct_start = _FNT_HEADER_SIZE[win_props.dfVersion]
    glyphs = []
    height = win_props.dfPixHeight
    entries = glyph_entry_array.from_bytes(data, ct_start)
    for ord, entry in enumerate(entries, win_props.dfFirstChar):
        # don't store empty glyphs but count them for ordinals
        if not entry.geWidth:
            continue
        bytewidth = ceildiv(entry.geWidth, 8)
        # transpose byte-columns to contiguous rows
        glyph_data = bytes(
            data[entry.geOffset + _col * height + _row]
            for _row in range(height)
            for _col in range(bytewidth)
        )
        glyph = Glyph.from_bytes(glyph_data, entry.geWidth).modify(codepoint=(ord,))
        glyphs.append(glyph)
    return glyphs


def _extract_vector_glyphs(data, win_props):
    """Read a vector character table."""
    n_chars = win_props.dfLastChar - win_props.dfFirstChar + 1
    ct_start = _FNT_HEADER_SIZE[win_props.dfVersion]
    if not win_props.dfPixWidth:
        # proportional font
        # always 2x2 bytes for prop. vector
        glyph_entry_array = _GLYPH_ENTRY_PVECTOR.array(n_chars+1)
    else:
        # fixed-width vector font
        glyph_entry_array = _GLYPH_ENTRY_1.array(n_chars+1)
    entries = glyph_entry_array.from_bytes(data, ct_start)
    offsets = tuple(_entry.geOffset for _entry in entries)
    if not win_props.dfPixWidth:
        widths = tuple(_entry.geWidth for _entry in entries)
    else:
        widths = tuple(win_props.dfPixWidth) * (n_chars+1)
    offset = win_props.dfBitsOffset
    glyphbytes = tuple(
        data[offset+_offset:offset+_next]
        for _offset, _next in zip(offsets, offsets[1:])
    )
    glyphdata = tuple(
        Props(codepoint=_cp, width=_w, code=_b)
        for _cp, (_w, _b) in enumerate(
            zip(widths, glyphbytes),
            win_props.dfFirstChar
        )
    )
    return _convert_vector_glyphs(glyphdata, win_props)


def _convert_vector_glyphs(glyphdata, win_props):
    """Convert a vector character table to paths."""
    glyphs = []
    for i, glyphrec in enumerate(glyphdata):
        # \x80 (-128) is the pen-up sentinel
        # all other bytes form signed int8 coordinate pairs
        code = le.int8.array(len(glyphrec.code)).from_bytes(glyphrec.code)
        it = iter(code)
        ink = StrokePath.LINE
        path = []
        for x in it:
            if x == -128:
                ink = StrokePath.MOVE
                continue
            try:
                y = next(it)
            except StopIteration:
                logging.warning('Vector glyph has truncated path definition')
                break
            path.append((ink, x, y))
            ink = StrokePath.LINE
        # Windows uses top-left coordinates
        path = StrokePath(path).flip().shift(0, win_props.dfAscent)
        glyphs.append(Glyph.from_path(
            path, advance_width=glyphrec.width, codepoint=glyphrec.codepoint
        ))
    return glyphs


def bytes_to_str(s, encoding='latin-1'):
    """Extract null-terminated string from bytes."""
    s, _, _ = s.partition(b'\0')
    return s.decode(encoding, errors='replace')

def _convert_win_props(data, win_props):
    """Convert WinFont properties to yaff properties."""
    version = win_props.dfVersion
    vector = win_props.dfType & 1
    if vector:
        logging.info('This is a vector font')
    logging.info('Windows FNT properties:')
    for key, value in win_props.__dict__.items():
        logging.info('    {}: {}'.format(key, value))
    properties = {
        'source_format': 'Windows FNT v{}.{}'.format(*divmod(version, 256)),
        'family': bytes_to_str(data[win_props.dfFace:]),
        'copyright': bytes_to_str(win_props.dfCopyright),
        'point_size': win_props.dfPoints,
        'slant': 'italic' if win_props.dfItalic else 'roman',
        # Windows dfAscent means distance between matrix top and baseline
        # and it calls the space where accents go the dfInternalLeading
        # which is specified to be 'inside the bounds set by dfPixHeight'
        'ascent': win_props.dfAscent - win_props.dfInternalLeading,
        # the dfPixHeight is the 'height of the character bitmap', i.e. our raster-size.y
        # and dfAscent is the distance between the raster top and the baseline,
        'descent': win_props.dfPixHeight - win_props.dfAscent,
        # dfExternalLeading is the 'amount of extra leading ... the application add between rows'
        'line_height': win_props.dfPixHeight + win_props.dfExternalLeading,
        'default_char': win_props.dfDefaultChar + win_props.dfFirstChar,
    }
    if not vector:
        # vector font determines shift-up (which only applies to raster) from path
        properties['shift_up'] = -properties['descent']
    if win_props.dfPixWidth:
        properties['spacing'] = 'character-cell'
    else:
        properties['spacing'] = 'proportional'
        # this can be extracted from the font - will be dropped if consistent
        # Windows documentation defines this as 'width of the character "X."'
        # for 1.0 system fonts, it is consistent with the advance width of LATIN CAPITAL LETTER X.
        # for 2.0+ system fonts, this appears to be set to the average advance width.
        # fontforge follows the "new" definition while mkwinfont follows the "old".
        # we'll make it depend on the version
        if version == 0x100:
            properties['cap_width'] = win_props.dfAvgWidth
        else:
            properties['average_width'] = win_props.dfAvgWidth
    # check prop/fixed flag
    if bool(win_props.dfPitchAndFamily & 1) == bool(win_props.dfPixWidth):
        logging.warning(
            'Inconsistent spacing properties: dfPixWidth=={} dfPitchAndFamily=={:04x}'.format(
                win_props.dfPixWidth, win_props.dfPitchAndFamily
            )
        )
    properties['dpi'] = (win_props.dfHorizRes, win_props.dfVertRes)
    deco = []
    if win_props.dfUnderline:
        deco.append('underline')
    if win_props.dfStrikeOut:
        deco.append('strikethrough')
    if deco:
        properties['decoration'] = ' '.join(deco)
    weight = win_props.dfWeight
    if weight:
        weight = max(100, min(900, weight))
        properties['weight'] = WEIGHT_MAP.get(round(weight, -2), None)
    charset = win_props.dfCharSet
    if charset in CHARSET_MAP:
        properties['encoding'] = CHARSET_MAP[charset]
    else:
        properties['windows.dfCharSet'] = str(charset)
    properties['style'] = _STYLE_MAP.get(win_props.dfPitchAndFamily & 0xff00, None)
    if win_props.dfBreakChar:
        properties['word_boundary'] = win_props.dfFirstChar + win_props.dfBreakChar
    properties['device'] = bytes_to_str(data[win_props.dfDevice:])
    # unparsed properties: dfMaxWidth - but this can be calculated from the matrices
    if version == 0x300:
        # https://github.com/letolabs/fontforge/blob/master/fontforge/winfonts.c
        # /* These fields are not present in 2.0 and are not meaningful in 3.0 */
        # /*  they are there for future expansion */
        # yet another prop/fixed flag
        if bool(win_props.dfFlags & _DFF_PROP) != (win_props.dfPixWidth == 0):
            logging.warning(
                'Inconsistent spacing properties: dfPixWidth=={} dfFlags=={:04x}'.format(
                    win_props.dfPixWidth, win_props.dfFlags
                )
            )
        if win_props.dfFlags & _DFF_COLORFONT:
            # https://web.archive.org/web/20120215123301/http://support.microsoft.com/kb/65123
            # NOTE: The only formats supported in Windows 3.0 will be DFF_FIXED and DFF_PROPORTIONAL.
            raise UnsupportedError('Colour fonts not supported')
        if win_props.dfFlags & _DFF_ABC:
            # https://ffenc.blogspot.com/2008/04/fnt-font-file-format.html
            # For Windows 3.00, the font-file header includes six new fields:
            # dFlags, dfAspace, dfBspace, dfCspace, dfColorPointer, and dfReserved1.
            # These fields are not used in Windows 3.00. To ensure compatibility with future
            # versions of Windows, these fields should be set to zero.
            raise UnsupportedError('ABC spacing properties not supported')
    return properties


##############################################################################
# windows .FNT writer

def _make_contiguous(font):
    """Fill out a contiguous range of glyphs."""
    # this width is 0 for proportional fonts
    blank = Glyph.blank(font.cell_size.x, font.raster_size.y)
    font = make_contiguous(font, supported_range=_FNT_RANGE, missing=blank)
    # add the guaranteed-blank glyph
    font = font.modify(font.glyphs + (blank,))
    return font


def create_fnt(font, version=0x200, vector=False):
    """Create .FNT from monobit font."""
    # ensure codepoint values are set, if possible
    # only include single-byte encoded glyphs
    # as firstChar and lastChar are byte-sized
    font = _make_contiguous(font)
    # bring to equal-height, equal-upshift, padded normal form
    font = font.equalise_horizontal()
    bitmaps, char_table, offset_bitmaps, byte_width = _convert_to_fnt_glyphs(
        font, version, vector
    )
    bitmap_size = sum(len(_b) for _b in bitmaps)
    win_props, header_ext, stringtable = _convert_to_fnt_props(
        font, version, vector,
        offset_bitmaps, bitmap_size, byte_width,
    )
    data = (
        bytes(win_props) + bytes(header_ext)
        + b''.join(char_table)
        + b''.join(bitmaps)
        + b''.join(stringtable)
    )
    return data


def _convert_to_fnt_props(
        font, version, vector,
        offset_bitmaps, bitmap_size, byte_width,
    ):
    """Convert font to FNT headers."""
    upshifts = set(_g.shift_up for _g in font.glyphs)
    shift_up, *remainder = upshifts
    assert not remainder
    # get lowest and highest codepoints (contiguous glyphs followed by blank)
    min_ord = font.glyphs[0].codepoint[0]
    max_ord = font.glyphs[-2].codepoint[0]
    # if encoding is compatible, use it; otherwise set to fallback value
    charset = CHARSET_REVERSE_MAP.get(font.encoding, _FALLBACK_CHARSET)
    default = font.get_glyph(font.default_char, missing='empty').codepoint
    if len(default) == 1:
        default_ord, = default
    else:
        default_ord = _FALLBACK_DEFAULT
    word_break = font.get_glyph(font.word_boundary, missing='empty').codepoint
    if len(word_break) == 1:
        break_ord, = word_break
    else:
        break_ord = _FALLBACK_BREAK
    if font.spacing == 'proportional' or vector:
        # low bit set for proportional
        pitch_and_family = 0x01 | _STYLE_REVERSE_MAP.get(font.style, 0)
        v3_flags = _DFF_PROPORTIONAL
        pix_width = 0
    else:
        # CHECK: is this really always set for fixed-pitch?
        pitch_and_family = _FF_MODERN
        v3_flags = _DFF_FIXED
        # x_width should equal average width
        pix_width = font.raster_size.x
    # add name and device strings
    face_name_offset = offset_bitmaps + bitmap_size
    face_name = font.family.encode('latin-1', 'replace') + b'\0'
    device_name_offset = face_name_offset + len(face_name)
    device_name = font.device.encode('latin-1', 'replace') + b'\0'
    file_size = device_name_offset + len(device_name)
    # set device name pointer to zero for 'generic font'
    if not device_name or device_name == b'\0':
        device_name_offset = 0
    try:
        weight = WEIGHT_REVERSE_MAP[font.weight]
    except KeyError:
        logging.warning(
            f'Weight `{font.weight}` not supported by Windows FNT resource format, '
            '`regular` will be used instead.'
        )
        weight = WEIGHT_REVERSE_MAP['regular']
    # create FNT file
    win_props = _FNT_HEADER(
        dfVersion=version,
        dfSize=file_size,
        dfCopyright=font.copyright.encode('ascii', 'replace')[:60].ljust(60, b'\0'),
        dfType=1 if vector else 0,
        dfPoints=int(font.point_size),
        dfVertRes=font.dpi.y,
        dfHorizRes=font.dpi.x,
        # Windows dfAscent means distance between matrix top and baseline
        # common shift_up is negative or zero in padded normal form
        dfAscent=font.raster_size.y + shift_up,
        #'ascent': win_props.dfAscent - win_props.dfInternalLeading,
        dfInternalLeading=font.raster_size.y + shift_up - font.ascent,
        #'line_height': win_props.dfPixHeight + win_props.dfExternalLeading,
        dfExternalLeading=font.line_height-font.raster_size.y,
        dfItalic=(font.slant in ('italic', 'oblique')),
        dfUnderline=('underline' in font.decoration),
        dfStrikeOut=('strikethrough' in font.decoration),
        dfWeight=WEIGHT_REVERSE_MAP.get(
            font.weight, WEIGHT_REVERSE_MAP['regular']
        ),
        dfCharSet=charset,
        dfPixWidth=pix_width,
        dfPixHeight=font.raster_size.y,
        dfPitchAndFamily=pitch_and_family,
        # for 2.0+, we use actual average advance here (like fontforge but unlike mkwinfont)
        dfAvgWidth=round(font.average_width),
        # max advance width
        dfMaxWidth=font.max_width,
        dfFirstChar=min_ord,
        dfLastChar=max_ord,
        dfDefaultChar=default_ord - min_ord,
        dfBreakChar=break_ord - min_ord,
        # strike width in bytes. not used for vector. (not used for v2.0+ ?)
        dfWidthBytes=byte_width,
        dfDevice=device_name_offset,
        dfFace=face_name_offset,
        dfBitsPointer=0, # used on loading
        dfBitsOffset=offset_bitmaps,
    )
    # version-specific header extension
    header_ext = _FNT_HEADER_EXT[version]()
    if version == 0x300:
        # all are zeroes (default) except the flags for v3
        header_ext.dfFlags = v3_flags
    stringtable = face_name, device_name
    return win_props, header_ext, stringtable


def _convert_to_fnt_glyphs(font, version, vector):
    """Convert glyphs to FNT bitmaps and offset tables."""
    if vector:
        upshifts = set(_g.shift_up for _g in font.glyphs)
        shift_up, *remainder = upshifts
        assert not remainder
        # vector glyph data
        # this should equal the dfAscent value
        win_ascent = font.raster_size.y + shift_up
        bitmaps = _convert_vector_glyphs_to_fnt(font.glyphs, win_ascent)
        byte_width = 0
    elif version == 0x100:
        strike = Raster.concatenate(*(_g.pixels for _g in font.glyphs))
        # spacer to ensure we'll be word-aligned (as_bytes will byte-align)
        spacer = Raster.blank(width=16-(strike.width%16), height=strike.height)
        strike = Raster.concatenate(strike, spacer)
        bitmaps = (strike.as_bytes(),)
        byte_width = ceildiv(strike.width, 8)
    else:
        # create the bitmaps
        bitmaps = (_glyph.as_bytes() for _glyph in font.glyphs)
        # bytewise transpose - .FNT stores as contiguous 8-pixel columns
        bitmaps = tuple(
            b''.join(
                _bm[_col::len(_bm)//_glyph.height]
                for _col in range(len(_bm)//_glyph.height)
            )
            for _glyph, _bm in zip(font.glyphs, bitmaps)
        )
        # not sure if this gets used for v2, as it isn't really useful there
        # using the logic from mkwinfont, max bytewidth aligned to multiple of 2
        byte_width = ceildiv(max(_g.width for _g in font.glyphs), 8)
        byte_width = align(byte_width, 1)
    if version == 0x100 and not vector:
        glyph_offsets = [0] + list(
            itertools.accumulate(_g.width for _g in font.glyphs)
        )
    else:
        glyph_offsets = [0] + list(
            itertools.accumulate(len(_bm) for _bm in bitmaps)
        )
    base_offset = _FNT_HEADER.size + _FNT_HEADER_EXT[version].size
    if not vector:
        glyph_entry = _GLYPH_ENTRY[version]
    else:
        glyph_entry = _GLYPH_ENTRY_PVECTOR
    # vector format and v1 do not include dfBitmapOffset in the table
    glyph_table_size = len(font.glyphs) * glyph_entry.size
    if version == 0x100 and font.spacing == 'character-cell':
        char_table = (b'',)
        offset_bitmaps = base_offset
    elif vector or version == 0x100:
        char_table = (
            bytes(glyph_entry(
                geWidth=_glyph.width,
                geOffset=_glyph_offset
            ))
            for _glyph, _glyph_offset in zip(font.glyphs, glyph_offsets)
        )
        offset_bitmaps = base_offset + glyph_table_size
    else:
        offset_bitmaps = base_offset + glyph_table_size
        char_table = (
            bytes(glyph_entry(
                geWidth=_glyph.width,
                geOffset=offset_bitmaps + _glyph_offset
            ))
            for _glyph, _glyph_offset in zip(font.glyphs, glyph_offsets)
        )
    return bitmaps, char_table, offset_bitmaps, byte_width


def _convert_vector_glyphs_to_fnt(glyphs, win_ascent):
    """Convert paths to a vector character table."""
    glyphdata = []
    for glyph in glyphs:
        path = glyph.path.shift(0, -win_ascent).flip()
        code = (
            (-128, _x, _y) if _ink == StrokePath.MOVE else (_x, _y)
            for _ink, _x, _y in path.as_moves()
        )
        code = tuple(_b for _tuple in code for _b in _tuple)
        code = le.int8.array(len(code))(*code)
        glyphdata.append(bytes(code))
    return glyphdata



"""
monobit.storage.fontformats.text.fd - mkwinfont .fd format

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string

from monobit.storage.base import loaders, savers
from monobit.core import Font, Glyph
from monobit.base import Props
from ..common import (
    WEIGHT_MAP, CHARSET_MAP, WEIGHT_REVERSE_MAP, CHARSET_REVERSE_MAP
)
from .draw import DrawComment, NonEmptyBlock, Empty, Unparsed, iter_blocks
from monobit.storage.utils.limitations import ensure_single


FD_KEYS = {
    'facename',
    'copyright',
    'height',
    'ascent',
    'pointsize',
    'weight',
    'charset',
    'italic',
    'underline',
    'strikeout',
}

FD_CHAR_KEYS = {
    'char',
    'width',
}

@loaders.register(
    name='mkwinfont',
    patterns=('*.fd',),
    text=True,
)
def load_mkwinfont(instream):
    """Load font from a mkwinfont .fd file."""
    properties, glyphs, comments = _read_mkwinfont(instream.text)
    return _convert_mkwinfont(properties, glyphs, comments)


@savers.register(linked=load_mkwinfont)
def save_mkwinfont(fonts, outstream):
    font = ensure_single(fonts)
    return _write_mkwinfont(font, outstream)



###############################################################################
# reader

def _read_mkwinfont(text_stream):
    """Read a mkwinfont file into a properties object."""
    # this will be a list of tuples
    font_props = {_k: None for _k in FD_KEYS}
    font_comments = []
    glyphs = []
    glyph_props = {}
    current_comment = []
    for block in iter_blocks(text_stream, (MWFGlyph, MWFProperties, MWFComment, Empty)):
        if isinstance(block, MWFComment):
            if not glyphs:
                font_comments.extend(current_comment)
                current_comment = []
            current_comment.append(block.get_value())
        elif isinstance(block, MWFProperties):
            properties = block.get_value()
            if glyphs or 'char' in properties:
                glyph_props.update(properties)
            else:
                font_props.update(properties)
                if current_comment:
                    font_comments.extend(current_comment)
                    current_comment = []
        elif isinstance(block, MWFGlyph):
            if not glyphs and not font_comments:
                font_comments.extend(current_comment)
                current_comment = []
            glyphs.append(Glyph(
                block.get_value(),
                inklevels='01',
                codepoint=glyph_props.pop('char', b''),
                comment='\n\n'.join(current_comment),
            ))
            glyph_props = {}
            current_comment = []
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    font_comments.extend(current_comment)
    return Props(**font_props), glyphs, font_comments

def _convert_mkwinfont(props, glyphs, comments):
    mb_props = dict(
        name=props.facename,
        copyright=props.copyright,
        # .fd doesn't store internal/external leading, so it gets included in descent
        descent=int(props.height)-int(props.ascent),
        ascent=props.ascent,
        point_size=props.pointsize,
        slant='italic' if props.italic else '',
        decoration=''.join((
            ('strikethrough ' if props.strikeout else ''),
            ('underline ' if props.underline else ''),
        )).strip(),
        weight=(
            None if props.weight is None else
            WEIGHT_MAP.get(round(max(100, min(900, int(props.weight))), -2), None)
        ),
        encoding=(
            None if props.charset is None else
            CHARSET_MAP.get(int(props.charset), None)
        ),
        comment='\n\n'.join(comments),
    )
    glyphs = (_g.modify(shift_up=-mb_props['descent']) for _g in glyphs)
    return Font(glyphs, **mb_props).label()


# mkwinfont block readers

class MWFGlyph(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in ('0', '1')

    def ends(self, line):
        return line[:1] not in ('0', '1')

    def get_value(self):
        return tuple(_l.strip() for _l in self.lines)


class MWFProperties(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in string.ascii_letters and ' ' in line

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        return dict(_l.partition(' ')[::2] for _l in self.lines)


class MWFComment(DrawComment):
    pass


###############################################################################
# writer

def _write_mkwinfont(font, outstream):
    outstream = outstream.text
    font = font.equalise_horizontal()
    props = _convert_to_mkwinfont_props(font)

    def _write_prop(key):
        if props[key] is not None and props[key] != '':
            outstream.write(f'{key} {props[key]}\n')

    if font.get_comment():
        outstream.write('# ' + '\n# '.join(font.get_comment().splitlines()) + '\n\n')
    _write_prop('facename')
    _write_prop('copyright')
    outstream.write('\n')
    _write_prop('height')
    _write_prop('ascent')
    if props['height'] != props['pointsize']:
        _write_prop('pointsize')
    if props['italic']:
        _write_prop('italic')
    if props['underline']:
        _write_prop('underline')
    if props['strikeout']:
        _write_prop('strikeout')
    if props['weight'] != 400:
        _write_prop('weight')
    outstream.write('\n')
    _write_prop('charset')
    outstream.write('\n')
    for codepoint in range(256):
        try:
            glyph = font.get_glyph(codepoint)
        except KeyError:
            pass
        else:
            outstream.write(f'char {codepoint}\n')
            outstream.write(f'width {glyph.advance_width}\n')
            outstream.write(glyph.as_text(inklevels='01'))
            outstream.write('\n')



def _convert_to_mkwinfont_props(font):
    return dict(
        facename=font.name,
        copyright=font.copyright,
        # .fd doesn't store internal/external leading, so we include everything
        height=font.line_height,
        ascent=font.ascent,
        pointsize=font.point_size,
        italic=font.slant == 'italic',
        strikeout='strikethrough' in font.decoration,
        underline='underline' in font.decoration,
        weight=WEIGHT_REVERSE_MAP.get(font.weight, None),
        charset=CHARSET_REVERSE_MAP.get(font.encoding, None),
    )



"""
monobit.storage.containerformats.yencode - yEncode format

(c) 2024 Rob Hagemans

with code from python3-yenc examples
https://github.com/oe-mirrors/python3-yenc/

> Copyright (C) 2003, 2011 Alessandro Duca <alessandro.duca@gmail.com>
>
> This library is free software; you can redistribute it and/or
> modify it under the terms of the GNU Lesser General Public
> License as published by the Free Software Foundation; either
> version 2.1 of the License, or (at your option) any later version.
>
> This library is distributed in the hope that it will be useful,
> but WITHOUT ANY WARRANTY; without even the implied warranty of
> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
> Lesser General Public License for more details.
>
> You should have received a copy of the GNU Lesser General Public
> License along with this library; if not, write to the Free Software
> Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
"""

import logging
from io import BytesIO
from pathlib import Path
import re
from binascii import crc32

from monobit.base import safe_import
yenc = safe_import('yenc')

from monobit.base import FileFormatError, Props
from ..magic import Sentinel
from ..base import containers
from ..containers import SerialContainer


if yenc:
    @containers.register(
        name='yenc',
        patterns=('*.yenc',),
        magic=(Sentinel(b'=ybegin '),),
        # last line starts with '=yend '
    )
    class YEncodeContainer(SerialContainer):
        """yEncode format container."""

        def decode(self, name):
            """
            Decode files from yEncode file. Accepts multiple, concatenated files.
            """
            return super().decode(name)

        def encode(self, name):
            """
            Encode files to yEncode file. Multiple files will be concatenated.
            """
            return super().encode(name)

        @classmethod
        def decode_all(cls, file_in):
            data = {}
            while file_in.peek(1):
                head_crc = trail_crc = None
                while True:
                    line = file_in.readline()
                    if line.startswith(b'=ybegin '):
                        metadata, name = line.split(b' name=')
                        name = name.rstrip().decode('latin-1')
                        header = _parse_header(metadata, '=ybegin ')
                        break
                    elif not line:
                        raise FileFormatError("No valid =ybegin header found.")
                if file_in.peek(7).startswith(b'=ypart '):
                    line = file_in.readline().rstrip()
                    part = _parse_header(line, '=ypart ')
                else:
                    part = Props(begin=0, end=header.size)
                with BytesIO() as file_out:
                    try:
                        dec, dec_crc = yenc.decode(file_in, file_out, int(header.size))
                    except yenc.Error as e:
                        raise FileFormatError(e) from e
                    garbage	= False
                    for line in file_in.read().splitlines():
                        if line.startswith(b'=yend '):
                            tailer = _parse_header(line, '=yend ')
                            break
                        elif not line:
                            continue
                        else:
                            garbage = True
                    else:
                        logging.warning("Couldn't find =yend trailer")
                    if garbage:
                        logging.warning("Garbage before =yend trailer")
                    if head_crc:
                        tmp_crc = header.crc.lower()
                    elif trail_crc:
                        tmp_crc = tailer.crc.lower()
                    else:
                        tmp_crc = dec_crc
                    if tmp_crc != dec_crc:
                        logging.warning(
                            "CRC32 mismatch: header: %s dec: %s", tmp_crc, dec_crc
                        )
                    if name not in data:
                        data[name] = bytearray(int(header.size))
                    data[name][int(part.begin):int(part.end)] = file_out.getvalue()
            return data

        @classmethod
        def _head(cls):
            return b''

        @classmethod
        def _tail(cls):
            return b''

        @classmethod
        def _separator(cls):
            return b'\r\n'

        @classmethod
        def _encode(cls, name, filedata, file_out):
            crc = b"%08x" % (0xFFFFFFFF & crc32(filedata))
            size = len(filedata)
            line = b"=ybegin line=128 size=%d crc32=%s name=%s\r\n"
            file_out.write(line % (size, crc, name.encode('latin-1')))
            with BytesIO(filedata) as file_in:
                try:
                    encoded, crc_out = yenc.encode(file_in, file_out, size)
                except Exception as e:
                    raise FileFormatError(e) from e
            line = b"=yend size=%d crc32=%s\r\n" % (encoded, crc_out.encode('ascii'))
            file_out.write(line)


    def _parse_header(line, prefix):
        try:
            line = line.decode('ascii').removeprefix(prefix)
            return Props(**dict(
                _kv.split('=')
                for _kv in line.split()
            ))
        except (ValueError, UnicodeError) as e:
            raise FileFormatError(f'Malformed ={prefix} header.') from e



"""
monobit.storage.fontformats.printshop - Broderbund's The Print Shop font files

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import count

from monobit.storage import loaders, savers
from monobit.core import Font, Glyph, Raster
from monobit.base import struct
from monobit.base.struct import little_endian as le
from monobit.base.binary import ceildiv

# offset 0
_HEADER = le.Struct(
    bbx_width='uint8',
    bbx_height='uint8',
    unused='uint16',
)
# offset 4
_WIDTHS = le.uint8.array(95)
# offset 99
_HEIGHTS = le.uint8.array(95)
# offset 194
_OFFSETS = le.uint16.array(95)
# offset 384, another 95 words, zeroed
# offset 574, bitmap header
_BITMAP_HEADER = le.Struct(
    unknown0='uint8',
    # maybe, but only sometimes fits
    maybe_baseline='uint8',
    unknown1='uint16',
    unknown2='uint16',
    # zero
    unused='uint8',
    name='9s',
    logo_bytewidth='uint8',
    logo_height='uint8',
    height='uint8',
    shortname='4s',
    unknown4='uint16',
    # unknown5==unknown2
    unknown5='uint16',
)
# logo bitmap starts at offset 601, followed by glyph


# not clear how baseline and interglyph spacing are determined
# the Print Shop uses negative bearings/kerning, but there clearly is
# no bearings or kerning table present in the font file
# so the hypothesis is that TPS does this algorithmically,
# moving the next glyph as far left as possible so that a certain amount
# of space remains between inked pixels

@loaders.register(
    name='printshop',
    patterns=('*.pnf', '*.psf'),
)
def load_printshop(instream):
    """
    Load a Broderbund The Print Shop font.
    """
    header = _HEADER.read_from(instream)
    logging.debug(header)
    widths = _WIDTHS.read_from(instream)
    heights = _HEIGHTS.read_from(instream)
    offsets = _OFFSETS.read_from(instream)
    # another block in the same format, usually (always?) zeroed out
    unknown = _OFFSETS.read_from(instream)
    logging.debug(unknown)
    bmp_header = _BITMAP_HEADER.read_from(instream)
    logging.debug(bmp_header)
    logobytes = instream.read(bmp_header.logo_bytewidth * bmp_header.logo_height)
    width = bmp_header.logo_bytewidth * 8
    logo = Glyph.from_bytes(
        logobytes, width=width, height=bmp_header.logo_height,
        tag='sample',
    )
    # ignoring the offsets table, assuming the glyphs are stored contiguously
    glyphs = tuple(
        Glyph.from_bytes(
            instream.read(ceildiv(_w, 8) * _h),
            width=_w, codepoint=_cp,
            shift_up=bmp_header.height-_h
        )
        for _w, _h, _cp in zip(widths, heights, count(0x20))
    )
    return Font(
        glyphs, source_format='The Print Shop',
        family=bmp_header.name.decode('latin-1'),
        # preserve the logo bitmap as a comment
        # storing it as a glyph would mess up the bounding box
        comment=logo.as_text()
    )
